"""discover_utils.data

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06th September 2023

The discover-utils data package defines data types for streams and annotations as well as providing respective data handlers to load and save data from and to various backends.
"""


"""Utility module to encode and decode specific classes in json

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    14.9.2023

"""

import json

from discover_utils.utils.ssi_xml_utils import Chain, ChainLink, Trainer, ModelIO, URI


class ModelIOEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for ModelIO objects.

    This encoder is used to serialize ModelIO objects to JSON format.

    Attributes:
        None

    """

    def default(self, obj):
        """
        Encodes a ModelIO object to JSON.

        Args:
            obj (ModelIO): The ModelIO object to encode.

        Returns:
            dict: A dictionary representation of the ModelIO object.

        """
        if isinstance(obj, ModelIO):
            return {"type": obj.io_type, "id": obj.io_id, "data": obj.io_data, "default_value": obj.io_default_value, "default_active": obj.io_default_active}
        return super().default(obj)

class ModelIODecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
    def object_hook(self, json_obj):
        if json_obj.get("type") and json_obj.get("id") and json_obj.get("data"):
            return ModelIO(json_obj["type"], json_obj["id"], json_obj["data"], json_obj["default_value"])
        else:
            raise ValueError("Invalid JSON format for ModelIO decoding.")


class URIEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for ModelIO objects.

    This encoder is used to serialize ModelIO objects to JSON format.

    Attributes:
        None

    """

    def default(self, obj):
        """
        Encodes a ModelIO object to JSON.

        Args:
            obj (ModelIO): The ModelIO object to encode.

        Returns:
            dict: A dictionary representation of the ModelIO object.

        """
        if isinstance(obj, URI):
            return {"id": obj.uri_id, "url": obj.uri_url, "hash": obj.uri_hash, "tar": obj.uri_tar}
        return super().default(obj)

class URIDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
    def object_hook(self, json_obj):
        if json_obj.get("id") and json_obj.get("url") and json_obj.get("hash"):
            return ModelIO(json_obj["id"], json_obj["url"], json_obj["hash"], json_obj["tar"])
        else:
            raise ValueError("Invalid JSON format for ModelIO decoding.")



class ChainLinkEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for ChainLink objects.

    This encoder is used to serialize ChainLink objects to JSON format.

    Attributes:
        None

    """

    def default(self, obj):
        """
        Encodes a ChainLink object to JSON.

        Args:
            obj (ChainLink): The ChainLink object to encode.

        Returns:
            dict: A dictionary representation of the ChainLink object.

        """
        if isinstance(obj, ChainLink):
            return {
                "create": obj.create,
                "script": obj.script,
                "optsstr": obj.optsstr,
                "syspath": obj.syspath,
                "tag": obj.tag,
                "multi_role_input": obj.multi_role_input,
            }
        return super().default(obj)


class ChainEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for Chain objects.

    This encoder is used to serialize Chain objects to JSON format.

    Attributes:
        None

    """

    def default(self, obj):
        """
        Encodes a Chain object to JSON.

        Args:
            obj (Chain): The Chain object to encode.

        Returns:
            dict: A dictionary representation of the Chain object.

        """
        if isinstance(obj, Chain):
            return {
                "meta_frame_step": obj.meta_frame_step,
                "meta_left_ctx": obj.meta_left_ctx,
                "meta_right_ctx": obj.meta_right_ctx,
                "meta_backend": obj.meta_backend,
                "meta_description": obj.meta_description,
                "meta_category": obj.meta_category,
                "meta_io": json.dumps(obj.meta_io, cls=ModelIOEncoder),
                "register": obj.register,
                "links": json.dumps(obj.links, cls=ChainLinkEncoder),
            }
        return super().default(obj)


class TrainerEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for Trainer objects.

    This encoder is used to serialize Trainer objects to JSON format.

    Attributes:
        None
    """

    def default(self, obj):
        """
        Encodes a Trainer object to JSON.

        Args:
            obj (Trainer): The Trainer object to encode.

        Returns:
            dict: A dictionary representation of the Trainer object.

        """
        if isinstance(obj, Trainer):
            return {
                "model_script_path": obj.model_script_path,
                "model_option_path": obj.model_option_path,
                "model_option_string": obj.model_optstr,
                "model_weights_path": obj.model_weights_path,
                "model_stream": obj.model_stream,
                "model_create": obj.model_create,
                "model_multirole_input": obj.model_multi_role_input,
                "users": obj.users,
                "classes": obj.classes,
                "streams": obj.streams,
                "register": obj.register,
                "info_trained": obj.info_trained,
                "meta_frame_step": obj.meta_frame_step,
                "meta_right_ctx": obj.meta_right_ctx,
                "meta_left_ctx": obj.meta_left_ctx,
                "meta_balance": obj.meta_balance,
                "meta_backend": obj.meta_backend,
                "meta_io": json.dumps(obj.meta_io, cls=ModelIOEncoder),
                "meta_uri": json.dumps(obj.meta_uri, cls=URIEncoder),
                "meta_description": obj.meta_description,
                "meta_is_iterable": obj.meta_is_iterable,
                "meta_enable_post_process": obj.meta_enable_post_process,
                "meta_category": obj.meta_category,
                "ssi_v": obj.ssi_v,
                "xml_version": obj.xml_version,
            }
        return super().default(obj)

if __name__ == '__main__':
    from pathlib import Path
    import os
    import dotenv
    dotenv.load_dotenv('../.env')
    data_dir = Path(os.getenv("DISCOVER_DATA_DIR"))
    trainer_in_fp = data_dir / "Test_Files" / "test.trainer"

    trainer = Trainer()
    trainer.load_from_file(trainer_in_fp)
    trainer_json = json.dumps(trainer, cls=TrainerEncoder)
    breakpoint()


"""
Discover-Utils Scripts

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06th September 2023

The Discover-Utils Scripts package is a collection of standalone scripts to perform specific machine learning tasks.
It includes a set of Python scripts that allow users to streamline the process of training machine learning models using Discover server modules.

Included Scripts:
    1. train.py - Training Script
    2. extract.py - Extract Script
    3. predict.py - Predict Script
"""


"""discover_utils.data.provider

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06th September 2023

The discover-utils data provider package implements different data providers to interact with multiple datastreams and annotations in a coherent way.
"""


"""Data Storage Class for session specific data
Author: Dominik Schiller <dominik.schiller@uni-a.de>
Date: 25.10.2023
"""
from pathlib import Path
from discover_utils.data.annotation import FreeAnnotation, FreeAnnotationScheme, ContinuousAnnotation, \
    ContinuousAnnotationScheme, DiscreteAnnotation, DiscreteAnnotationScheme
from discover_utils.utils import path_utils
from discover_utils.data.data import Data
from discover_utils.data.handler import (
    file_handler,
    nova_db_handler,
    url_handler,
    request_handler,
)
from discover_utils.data.stream import SSIStream, Video, Audio, Stream
from  discover_utils.data.static import Image, Text
from discover_utils.utils.request_utils import Origin, SuperType, SubType, parse_src_tag, data_description_to_string, \
    infer_dtype


class SessionManager:
    """
    Class to aggregate and manage interrelated incoming and outgoing datastreams belonging to a single session (e.g multimodal data from the same recording).

    Attributes:
        input_data (dict, optional):  Annotation or Stream data that can be processed by a module.
        dataset (str, optional): The dataset or category the session belongs to.
        data_description (list[dict[str, str]], optional): List of data descriptions. Defaults to None. The dictionary should have the following mandatory fields:
        source_context (dict[str, dict]) : Dict of parameters that are need to interact with a source. E.g. database credentials or data directories. Must match constructor arguments of the respective data handler.
        session (str, optional): The name or title of the session.
        duration (int, optional): The duration of the session in minutes.
        location (str, optional): The location or venue of the session.
        language (str, optional): The language used in the session.
        date (datetime, optional): The date and time of the session.
        is_valid (bool, optional): Whether the session is considered valid.
        extra_data (dict, optional): Additional stream or annotation information for the session. Might only contain meta information.
        output_data_templates (dict, optional): Empty data objects that contain meta information for outputs. Modules can fill the data here.

    Args:
        dataset (str, optional): The dataset or category the session belongs to. Must match NovaDB entries if 'db' is
        data_description (list[dict[str, str]], optional): List of data descriptions. Defaults to None. The dictionary should have the following mandatory fields:
            ``"id"``:
                Unique id to map the data to a given input / output.
            ``"name"``:
                Output name for streams
            ``"type"``:
                IO type of the data. Either "input" or "output"
            ``"src"``
                The source and datatype to load the data from separated by ':' . Source corresponds to any Origin.value and datatype has the format '<SuperType.value>:<SubType.value>:str'
                E.g. 'db:stream:video:specification'

            In addition, each entry should provide the information that is need to identify the exact input and output targets.
            Must match the input parameters of the respective data handlers save and load function (dataset and session are already specified as properties).
            E.g. to load an annotation from NovaDB we use the data.handler.nova_db_handler.AnnotationHandler we need the following additional fields.
                ``"scheme"``
                    The scheme name of the annotations to load. Only necessary when loading annotations from the database.
                ``"annotator"``
                    The annotator of the annotations to load. Only necessary when loading annotations from the database.
                ``"role"``
                    The role to which the data belongs. Only necessary when accessing data from the database.
            To load a stream file from disk using data.handler.file_handler.FileHandler we only need a filepath
            ``"uri"``
                The filepath from which to load the data from. Only necessary when loading files from disk.

        session (str, optional): The name or title of the session.
        source_context (dict[str, dict]) : List of parameters that are need to interact with a source. E.g. database credentials or data directories. Must match constructor arguments of the respective data handler.
        duration (int, optional): The duration of the session in milliseconds.
        location (str, optional): The location or venue of the session.
        language (str, optional): The language used in the session.
        date (datetime, optional): The date and time of the session.
        is_valid (bool, optional): Whether the session is considered valid.
        input_data (dict, optional): Annotation or Stream data that can be processed by a module.
        extra_data (dict, optional): Additional stream or annotation information for the session. Might only contain meta information.
        output_data_templates (dict, optional): Empty data objects that contain meta information for outputs. Modules can fill the data here.
    """

    def __init__(
            self,
            dataset: str = None,
            data_description: list[dict[str, str]] = None,
            session: str = None,
            source_context: dict[str, dict] = None,
            input_data: dict = None,
            extra_data: dict = None,
            output_data_templates: dict = None,
            video_backend: file_handler.VideoBackend = file_handler.VideoBackend.IMAGEIO,
    ):
        self.dataset = dataset
        self.data_description = data_description
        self.session = session
        self.input_data = {} if input_data is None else input_data
        self.extra_data = {} if extra_data is None else extra_data
        self.output_data_templates = (
            {} if output_data_templates is None else output_data_templates
        )
        self.source_context = {}
        if source_context is not None:
            for src, context in source_context.items():
                src_ = Origin(src)
                self.add_source_context(src_, context)

        self.video_backend = video_backend

    def add_source_context(self, source: Origin, context: dict):
        """Add all parameters that are necessary to initialize source specific data handler for reading and writing data objects."""
        self.source_context[source] = context

    def _update_data_description(self, data_description=None):
        if data_description is not None:
            self.data_description = data_description
        return self.data_description

    def load(self, data_description=None):
        """
        Args:
            data_description (list[dict[str, str]], optional): List of data descriptions. Defaults to None. The dictionary should have the following mandatory fields:
            ``"id"``:
                Unique id to map the data to a given input / output.
            ``"name"``:
                Output name for streams
            ``"type"``:
                IO type of the data. Either "input" or "output"
            ``"src"``
                The source and datatype to load the data from separated by ':' . Source corresponds to any Origin.value and datatype has the format '<SuperType.value>:<SubType.value>:str'
                E.g. 'db:stream:video:specification'

            In addition, each entry should provide the information that is need to identify the exact input and output targets.
            Must match the input parameters of the respective data handlers save and load function (dataset and session are already specified as properties).
            E.g. to load an annotation from NovaDB we use the data.handler.nova_db_handler.AnnotationHandler we need the following additional fields.
                ``"scheme"``
                    The scheme name of the annotations to load. Only necessary when loading annotations from the database.
                ``"annotator"``
                    The annotator of the annotations to load. Only necessary when loading annotations from the database.
                ``"role"``
                    The role to which the data belongs. Only necessary when accessing data from the database.
            To load a stream file from disk using data.handler.file_handler.FileHandler we only need a filepath
            ``"fp"``
                The filepath from which to load the data from. Only necessary when loading files from disk.

        Returns:

        """

        data_description = self._update_data_description(data_description)
        if data_description is None:
            raise ValueError(
                "Data description is empty. Either pass a data description to load() or set it as class attribute."
            )

        for desc in data_description:
            src, super_dtype, sub_dtype, specific_dtype = parse_src_tag(desc)

            header_only = False
            if desc.get("type") == "input":
                io_dst = self.input_data
            elif desc.get("type") == "output":
                io_dst = self.output_data_templates
                header_only = True
            else:
                io_dst = self.extra_data

            data_id = data_description_to_string(desc)
            data = None

            if src in [Origin.DB] and not src in self.source_context.keys():
                raise ValueError(
                    f"Missing context information source {src}. Call add_source_context() first."
                )
            try:
                # DATABASE
                if src == Origin.DB:
                    ctx = self.source_context[src]
                    if super_dtype == SuperType.ANNO:
                        handler = nova_db_handler.AnnotationHandler(**ctx)
                        data = handler.load(
                            dataset=self.dataset,
                            session=self.session,
                            scheme=desc["scheme"],
                            annotator=desc["annotator"],
                            role=desc["role"],
                            header_only=header_only
                        )
                    elif super_dtype == SuperType.STREAM:
                        handler = nova_db_handler.StreamHandler(**ctx)
                        data = handler.load(
                            dataset=self.dataset,
                            session=self.session,
                            name=desc["name"],
                            role=desc["role"],
                            header_only=header_only
                        )
                # FILE
                elif src == Origin.FILE:
                    fp = Path(path_utils.get_tmp_dir()) / desc['uri']
                    handler = file_handler.FileHandler(video_backend=self.video_backend)
                    data = handler.load(fp=fp, header_only=header_only)
                # URL
                elif src == Origin.URL:
                    handler = url_handler.URLHandler()
                    data = handler.load(url=desc["uri"])
                # REQUEST
                elif src == Origin.REQUEST:
                    target_dtype = infer_dtype(super_dtype, sub_dtype)
                    handler = request_handler.RequestHandler()
                    data = handler.load(data=desc.get("data"), dtype=target_dtype, header_only=header_only)

            except FileNotFoundError as e:
                # Only raise file not found error if stream is requested as input
                if not header_only:
                    raise e

                # Create empty data objects with known params
                else:
                    if super_dtype == SuperType.STREAM:
                        if sub_dtype == SubType.SSIStream:
                            data_cls = SSIStream
                        elif sub_dtype == SubType.VIDEO:
                            data_cls = Video
                        elif sub_dtype == SubType.AUDIO:
                            data_cls = Audio
                        else:
                            data_cls = Stream
                        data = data_cls(
                            None,
                            name=desc.get("name"),
                            role=desc.get("role"),
                            sample_rate=1,
                            dataset=self.dataset,
                            session=self.session,
                        )
                    elif super_dtype == SuperType.IMAGE:
                        data_cls = Image
                        data = data_cls(
                            None,
                            name=desc.get("name"),
                            role=desc.get("role"),
                            dataset=self.dataset,
                            session=self.session,
                        )
                    elif super_dtype == SuperType.TEXT:
                        data_cls = Text
                        data = data_cls(
                            None,
                            name=desc.get("name"),
                            role=desc.get("role"),
                            dataset=self.dataset,
                            session=self.session,
                        )
                    elif super_dtype == SuperType.ANNO:
                        if sub_dtype is None:
                            sub_dtype = SubType.FREE
                        print(
                            f'No predefined annotation scheme available. Creating generic {sub_dtype.value} template annotation for {desc}')
                        if sub_dtype == SubType.FREE:
                            data = FreeAnnotation(scheme=FreeAnnotationScheme(name='generic'), data=None)
                        elif sub_dtype == SubType.CONTINUOUS:
                            data = ContinuousAnnotation(
                                scheme=ContinuousAnnotationScheme(name='generic', sample_rate=1, min_val=0, max_val=1),
                                data=None)
                        elif sub_dtype == SubType.DISCRETE:
                            data = DiscreteAnnotation(scheme=DiscreteAnnotationScheme(name='generic', classes={'1' : 'class_one', '2': 'class_two'}))
                        else:
                            raise ValueError(
                                f"Can\'t create template for {desc} because no scheme information is available.")

                    else:
                        # Todo Handle other cases where no header might be loaded
                        data = Data()

            io_dst[data_id] = data

    def save(self, data_description=None, overwrite=True):
        """
        Args:
          data_description (list[dict[str, str]], optional): List of data descriptions. Defaults to None. The dictionary should have the following mandatory fields:
          ``"id"``:
              Unique id to map the data to a given input / output.
          ``"name"``:
              Output name for streams
          ``"type"``:
              IO type of the data. Either "input" or "output"
          ``"src"``
              The source and datatype to load the data from separated by ':' . Source is of typ Source.value and datatype of type DType.value
              E.g. 'db:anno'

          In addition, each entry should provide the information that is need to identify the exact input and output targets.
          Must match the input parameters of the respective data handlers save and load function (dataset and session are already specified as properties).
          E.g. to load an annotation from NovaDB we use the data.handler.nova_db_handler.AnnotationHandler we need the following additional fields.
              ``"scheme"``
                  The scheme name of the annotations to load. Only necessary when loading annotations from the database.
              ``"annotator"``
                  The annotator of the annotations to load. Only necessary when loading annotations from the database.
              ``"role"``
                  The role to which the data belongs. Only necessary when accessing data from the database.
          To load a stream file from disk using data.handler.file_handler.FileHandler we only need a filepath
          ``"fp"``
              The filepath from which to load the data from. Only necessary when loading files from disk.

        Returns:
        """

        data_description = self._update_data_description(data_description)
        if data_description is None:
            raise ValueError(
                "Data description is empty. Either pass a data description to save() or set it as class attribute."
            )

        success = False

        for desc in data_description:

            # Do not write the output if active flag is not set
            if not desc.get('active', True):
                continue

            src, super_dtype, sub_dtype, specific_dtype = parse_src_tag(desc)

            if not desc.get("type") == "output":
                continue

            data_id = data_description_to_string(desc)

            if src in [Origin.DB] and not src in self.source_context.keys():
                raise ValueError(
                    f"Missing context information source {src}. Call add_source_context() first."
                )

            data = self.output_data_templates[data_id]
            if src == Origin.DB:
                ctx = self.source_context[src]
                if super_dtype == SuperType.ANNO:
                    role = desc.get('role')
                    scheme = desc.get('scheme')
                    annotator = desc.get('annotator')
                    handler = nova_db_handler.AnnotationHandler(**ctx)
                    success = handler.save(dataset=self.dataset, session=self.session, annotation=data, overwrite=overwrite, role=role, annotator=annotator, scheme=scheme)
                elif super_dtype == SuperType.STREAM:
                    name = desc.get('name')
                    role = desc.get('role')
                    handler = nova_db_handler.StreamHandler(**ctx)
                    success = handler.save(dataset=self.dataset, session=self.session, stream=data, role=role, name=name)
            elif src == Origin.FILE:
                handler = file_handler.FileHandler()
                success = handler.save(data=data, fp=Path(desc["uri"]))
            elif src == Origin.URL:
                raise NotImplementedError
            elif src == Origin.REQUEST:
                rq = self.source_context.get(Origin.REQUEST)
                shared_dir = rq.get('shared_dir')
                job_id = rq.get('job_id')
                handler = request_handler.RequestHandler()
                handler.save(data=data, shared_dir=shared_dir, job_id=job_id, dataset=self.dataset,
                             session=self.session)

        return success


class DatasetManager:
    def __init__(
            self, data_description: list[dict[str, str]], source_context: dict = None, dataset: str = None,
            session_names: list = None, video_backend: file_handler.VideoBackend = file_handler.VideoBackend.IMAGEIO
    ):
        self.dataset = dataset
        self.data_description = data_description
        self.session_names = session_names
        self.source_ctx = source_context
        self.sessions = {}
        self.video_backend = video_backend
        self._init_sessions()

    def _init_sessions(self):
        db_required = any([parse_src_tag(dd)[0] == Origin.DB for dd in self.data_description])

        # Load session information from database
        if db_required:
            sh = nova_db_handler.SessionHandler(**self.source_ctx["db"])
            sessions = sh.load(self.dataset, self.session_names)
            for sess in sessions:
                sm = SessionManager(
                    self.dataset, self.data_description, sess.name, self.source_ctx, video_backend = self.video_backend
                )
                self.sessions[sess.name] = {"manager": sm, "info": sess}
            self.session_names = list(self.sessions.keys())

        else:
            if not self.session_names:
                self.session_names = ['dummy_session']

            if self.dataset is None:
                self.dataset = 'dummy_dataset'

            for session in self.session_names:
                sm = SessionManager(
                    self.dataset, self.data_description, session, self.source_ctx, video_backend = self.video_backend
                )
                self.sessions[session] = {"manager": sm, "info": None}

    def load_session(self, session_name):
        self.sessions[session_name]["manager"].load(self.data_description)

    def save_session(self, session_name):
        self.sessions[session_name]["manager"].save(self.data_description)

    def load(self):
        for session in self.session_names:
            self.load_session(session)

    def save(self):
        for session in self.session_names:
            self.save_session(session)


# class NovaDatasetManager(DatasetManager):
#
#     def _init_sessions(self):
#         sh = nova_db_handler.SessionHandler(**self.source_ctx["db"])
#         sessions = sh.load(self.dataset, self.session_names)
#         for session_info in sessions:
#             sm = SessionManager(
#                 self.dataset, self.data_description, session_info.name, self.source_ctx
#             )
#             self.sessions[session_info.name] = {"manager": sm, "info": session_info}


if __name__ == "__main__":
    from dotenv import load_dotenv
    import os
    import dotenv
    dotenv.load_dotenv()
    DISCOVER_TEST_FILE_DIR = Path(os.getenv("DISCOVER_DATA_DIR"))
    DISCOVER_OUR_DIR = Path(os.getenv("DISCOVER_TEST_DIR"))

    IP = os.getenv("NOVA_IP", "")
    PORT = int(os.getenv("NOVA_PORT", 0))
    USER = os.getenv("NOVA_USER", "")
    PASSWORD = os.getenv("NOVA_PASSWORD", "")
    DATA_DIR = Path(os.getenv("NOVA_DATA_DIR", None))

    DATASET = os.getenv("DISCOVER_ITERATOR_TEST_DATASET")
    SESSIONS = [os.getenv("DISCOVER_ITERATOR_TEST_SESSION")]
    SCHEME = os.getenv("DISCOVER_ITERATOR_TEST_SCHEME")
    ANNOTATOR = os.getenv("DISCOVER_ITERATOR_TEST_ANNOTATOR")
    ROLE = os.getenv("DISCOVER_ITERATOR_TEST_ROLE")
    FEATURE_STREAM = os.getenv("DISCOVER_ITERATOR_TEST_STREAM")


    #dataset = "test"
    #sessions = ["01_AffWild2_video1"]

    annotation = {
        "src": "db:annotation",
        "scheme": SCHEME,
        "type": "input",
        "id": "annotation",
        "annotator": ANNOTATOR,
        "role": ROLE,
    }

    # stream = {
    #     "src": "db:stream",
    #     "type": "input",
    #     "id": "featurestream",
    #     "role": ROLE,
    #     "name": FEATURE_STREAM,
    # }

    # file = {
    #     "src": "file:stream:Audio",
    #     "type": "input",
    #     "id": "file",
    #     "uri": DISCOVER_TEST_FILE_DIR/"test_audio.wav",
    # }

    # request = {
    #     "src": "request:text:video:dum:dum:dum",
    #     "type": "input",
    #     "id": "reqeust_test",
    #     "data": "this is some input"
    # }

    ctx = {
        "db": {
            "db_host": IP,
            "db_port": PORT,
            "db_user": USER,
            "db_password": PASSWORD,
            "data_dir": DATA_DIR,
        },
    }

    annotation_out = {
        "src": "file:annotation",
        "type": "output",
        "id": "annotation_out",
        "uri": DISCOVER_OUR_DIR/"test_output.annotation",
    }

    dsm = DatasetManager(
        dataset=DATASET, session_names=SESSIONS, data_description=[annotation, annotation_out], source_context=ctx)

    # Loading session and set as output
    session_manager = dsm.sessions['01_AffWild2_video1']['manager']
    session_manager.load()
    session_manager.output_data_templates["annotation_out"] = session_manager.input_data["annotation"]

    dsm.save()

    breakpoint()



"""Standalone script for general processing

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    20.09.2023

This script performs general data processing to extract either annotations to NOVA-Database or streams to disk using a provided DISCOVER module for inference.

.. argparse::
   :module: discover_utils.scripts.process
   :func: parser
   :prog: du-process

Returns:
    None

Example:
    >>> du-process --dataset "test" --db_host "127.0.0.1" --db_port "37317" --db_user "my_user" --db_password "my_password" --trainer_file_path "test\\test_predict.trainer" --sessions "[\"test_session_1\", \"test_session_2\"]" --data "[{\"src\": \"db:anno\", \"scheme\": \"transcript\", \"annotator\": \"test\", \"role\": \"testrole\"}]" --frame_size "0" --left_context "0" --right_context "0" --job_i_d "test_job" --opt_str "num_speakers=2;speaker_ids=testrole,testrole2" --cml_dir "./cml" --data_dir "./data" --log_dir "./log" --cache_dir "./cache" --tmp_dir "./tmp"
"""

import argparse
import sys
import os
import shutil
import traceback
from importlib.machinery import SourceFileLoader
from pathlib import Path, PureWindowsPath
from typing import Union, Type

import discover_utils.data.handler.file_handler
from discover_utils.data.provider.data_manager import DatasetManager, SessionManager
from discover_utils.data.provider.dataset_iterator import DatasetIterator
from discover_utils.interfaces.server_module import Predictor, Extractor
from discover_utils.scripts.parsers import (
    dm_parser,
    nova_db_parser,
    request_parser,
    nova_iterator_parser,
    nova_server_module_parser,
    io_parser
)
from discover_utils.utils import ssi_xml_utils, string_utils
from discover_utils.utils.string_utils import string_to_bool, parse_time_string_to_ms
from discover_utils.utils.anno_utils import pack_remove
from discover_utils.data.annotation import DiscreteAnnotation

# Main parser for predict specific options
parser = argparse.ArgumentParser(
    description="Use a provided DISCOVER module for inference and save results to NOVA-DB",
    parents=[dm_parser,
             nova_db_parser,
             request_parser,
             nova_iterator_parser,
             nova_server_module_parser,
             io_parser],
    fromfile_prefix_chars='@',
)
parser.add_argument(
    "--trainer_file_path",
    type=str,
    required=True,
    help="Path to the trainer file using Windows UNC-Style",
)

parser.add_argument(
    "--anno_min_dur",
    type=str,
    required=False,
    default=-999999999,
    help="Minimum duration for discrete annotations in either seconds (float or 's'-suffix) or milliseconds (int or 'ms'-suffix)",
)

parser.add_argument(
    "--anno_min_gap",
    type=str,
    required=False,
    default=-999999999,
    help="Minimum gap between labels of the same class to be considered separate labels. Specified in either seconds (float or 's'-suffix) or milliseconds (int or 'ms'-suffix)",
)


def main(args):

    process_args, _ = parser.parse_known_args(args)

    # Create argument groups
    db_args, _ = nova_db_parser.parse_known_args(args)
    req_args, _ = request_parser.parse_known_args(args)
    dm_args, _ = dm_parser.parse_known_args(args)
    iter_args, _ = nova_iterator_parser.parse_known_args(args)
    module_args, _ = nova_server_module_parser.parse_known_args(args)
    io_args, _ = io_parser.parse_known_args(args)

    # Set environment variables
    os.environ['CACHE_DIR'] = module_args.cache_dir
    os.environ['TMP_DIR'] = module_args.tmp_dir

    caught_ex = False

    # Preprocess options
    anno_min_dur = parse_time_string_to_ms(process_args.anno_min_dur)
    anno_min_gap = parse_time_string_to_ms(process_args.anno_min_gap)

    # Load trainer
    trainer = ssi_xml_utils.Trainer()
    trainer_file_path = Path(module_args.cml_dir).joinpath(
        PureWindowsPath(process_args.trainer_file_path)
    )
    if not trainer_file_path.is_file():
        raise FileNotFoundError(f"Trainer file not available: {trainer_file_path}")
    else:
        trainer.load_from_file(trainer_file_path)
        print("Trainer successfully loaded.")

    # Load module
    if not trainer.model_script_path:
        raise ValueError('Trainer has no attribute "script" in model tag.')

    model_script_path = (
            trainer_file_path.parent / PureWindowsPath(trainer.model_script_path)
    ).resolve()
    source = SourceFileLoader(
        "ns_cl_" + model_script_path.stem, str(model_script_path)
    ).load_module()
    print(f"Trainer module {Path(model_script_path).name} loaded")

    opts = module_args.options
    if module_args.options is None:
        opts = string_utils.parse_nova_option_string(process_args.opt_str)
        print('Option --opt_str is deprecated. Use --options in the future.')

    processor_class: Union[Type[Predictor], Type[Extractor]] = getattr(
        source, trainer.model_create
    )
    processor = processor_class(model_io=trainer.meta_io, opts=opts, trainer=trainer)
    print(f"Model {trainer.model_create} created")


    # Build data loaders
    ctx = {
        'db' : {
            **vars(db_args)
        },
        'request' : {
            **vars(req_args)
        }
    }

    # Clear output directory for job id
    shared_dir = ctx['request'].get('shared_dir')
    job_id = ctx['request'].get('job_id')
    if shared_dir and job_id:
        output_dir = Path(shared_dir) / job_id
        if output_dir.exists():
            if output_dir.is_dir():
                shutil.rmtree(output_dir)
        if output_dir.is_file():
            output_dir.unlink()

    single_session_data_provider = []

    # Create one dataset per session
    video_backend = discover_utils.data.handler.file_handler.VideoBackend[io_args.video_backend]
    for session in dm_args.sessions:
        is_iterable = string_to_bool(trainer.meta_is_iterable)
        if is_iterable:
            data_provider = DatasetIterator(dataset=dm_args.dataset, data_description=dm_args.data, source_context=ctx, session_names=[session], video_backend= video_backend, **vars(iter_args))
        else:
            data_provider = DatasetManager(dataset=dm_args.dataset, data_description=dm_args.data, source_context=ctx, session_names=[session],  video_backend= video_backend)

        single_session_data_provider.append(data_provider)
    print("Data managers initialized")

    # Iterate over all sessions
    total_sessions = len(single_session_data_provider)
    for session_idx, provider in enumerate(single_session_data_provider):
        session = provider.session_names[0]
        #data_provider = provider
        try:
            if isinstance(provider, DatasetManager):
                provider.load()

            # Data processing with progress information
            print(f"Processing session ({session_idx + 1}/{total_sessions}): {session}...")

            # Update progress in DISCOVER via HTTP/HTTPS
            try:
                job_key = os.environ.get('DISCOVER_JOB_KEY')
                discover_host = os.environ.get('DISCOVER_HOST', 'localhost')
                discover_port = os.environ.get('DISCOVER_PORT', '8080')
                use_tls = os.environ.get('DISCOVER_USE_TLS', 'false').lower() == 'true'

                # Convert 0.0.0.0 to localhost for client requests
                if discover_host == '0.0.0.0':
                    discover_host = 'localhost'

                if job_key:
                    import requests
                    import warnings
                    from urllib3.exceptions import InsecureRequestWarning

                    # Suppress InsecureRequestWarning for self-signed certificates
                    warnings.filterwarnings('ignore', category=InsecureRequestWarning)

                    protocol = 'https' if use_tls else 'http'
                    url = f"{protocol}://{discover_host}:{discover_port}/api/update_progress"
                    payload = {
                        "job_key": job_key,
                        "progress": f"{session_idx + 1}/{total_sessions}"
                    }
                    # Disable SSL verification for self-signed certificates
                    requests.post(url, json=payload, timeout=1, verify=False)
            except Exception:
                pass  # Fail silently if progress update fails

            data_processed = processor.process_data(provider)
            data_output = processor.to_output(data_processed)

            # Data Saving
            session_manager : SessionManager
            session_manager = provider.sessions[session]['manager']
            for io_id, data_object in data_output.items():
                if isinstance(data_object, DiscreteAnnotation):
                    data_object.data = pack_remove(data_object.data, min_gap=anno_min_gap, min_dur=anno_min_dur)
                session_manager.output_data_templates[io_id] = data_object

            provider.save()
            print(f"Completed session ({session_idx + 1}/{total_sessions}): {session}...")

        except Exception as e:
            traceback.print_exc()
            print(
                f"\tProcessor exited with error: '{str(e)}'.\nContinuing with next session ({session_idx + 1}/{total_sessions})."
            )
            caught_ex = True
            continue

    print("Processing completed!")
    if caught_ex:
        print(
            "Processing job encountered errors for some sessions. Check logs for details."
        )
        exit(1)

# Entry point for du-process
def cl_main():
    main(sys.argv[1:])

# Entry point for python
if __name__ == "__main__":
    main(sys.argv[1:])



import decord
import numpy as np
from decord import cpu
from pims import FramesSequence, Frame

class DecordReader(FramesSequence):

    def __init__(self, filename):
        self.filename = filename
        self.reader = decord.VideoReader(filename, ctx=cpu(0))
        self._len = self.reader._num_frame
        _f = self.reader[0]
        self.reader.seek(0)
        self._dtype = np.dtype(_f.dtype)
        self._frame_shape = _f.shape[:2]

    def get_frame(self, i):
        # Access the data you need and get it into a numpy array.
        # Then return a Frame like so:
        return Frame(self.reader[i].asnumpy(), frame_no=i)

    def __len__(self):
        return self._len

    @property
    def frame_shape(self):
        return self._frame_shape

    @property
    def pixel_type(self):
        return self._dtype


"""General logic for predicting annotations to the nova database

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06.09.2023

"""

def _main():
    raise NotImplementedError

if __name__ == '__main__':
    _main()


""" Module for handling MongoDB data operations related to annotations and streams.

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    18.8.2023
"""

import os
import warnings
from datetime import datetime
from pathlib import Path
from typing import Union

import numpy as np
from bson.objectid import ObjectId
from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError
from pymongo.results import InsertOneResult, UpdateResult

from discover_utils.data.annotation import (
    Annotation,
    DiscreteAnnotation,
    DiscreteAnnotationScheme,
    ContinuousAnnotation,
    ContinuousAnnotationScheme,
    FreeAnnotation,
    FreeAnnotationScheme,
)
from discover_utils.data.handler.file_handler import FileHandler
from discover_utils.data.handler.ihandler import IHandler
from discover_utils.data.stream import Stream, SSIStream, StreamMetaData
from discover_utils.utils.anno_utils import (
    convert_ssi_to_label_dtype,
    convert_label_to_ssi_dtype,
    resample,
    remove_label
)
from discover_utils.utils.type_definitions import SSILabelDType, SchemeType

ANNOTATOR_COLLECTION = "Annotators"
SCHEME_COLLECTION = "Schemes"
STREAM_COLLECTION = "Streams"
ROLE_COLLECTION = "Roles"
ANNOTATION_COLLECTION = "Annotations"
SESSION_COLLECTION = "Sessions"
ANNOTATION_DATA_COLLECTION = "AnnotationData"


# METADATA
class NovaDBMetaData:
    """
    Metadata for MongoDB connection.

    Attributes:
        ip (str, optional): IP address of the MongoDB server.
        port (int, optional): Port number of the MongoDB server.
        user (str, optional): Username for authentication.
        dataset (str, optional): Name of the dataset.
    """

    def __init__(
            self, ip: str = None, port: int = None, user: str = None, dataset: str = None
    ):
        self.ip = ip
        self.port = port
        self.user = user
        self.dataset = dataset


class NovaDBAnnotationMetaData(NovaDBMetaData):
    """
    Metadata for MongoDB annotations.

    Attributes:
        is_locked (bool, optional): Indicates if the annotation is locked.
        is_finished (bool, optional): Indicates if the annotation is finished.
        last_update (bool, optional): Timestamp of the last update.
        annotation_document_id (ObjectId, optional): ID of the annotation document.
        data_document_id (ObjectId, optional): ID of the associated data document.
        **kwargs: Arbitrary keyword arguments.
    """

    def __init__(
            self,
            is_locked: bool = None,
            is_finished: bool = None,
            last_update: bool = None,
            annotation_document_id: ObjectId = None,
            data_document_id: ObjectId = None,
            **kwargs,
    ):
        super().__init__(**kwargs)
        self.is_locked = is_locked
        self.is_finished = is_finished
        self.last_update = last_update
        self.annotation_document_id = annotation_document_id
        self.data_document_id = data_document_id


class NovaDBStreamMetaData(NovaDBMetaData):
    """
    Metadata for MongoDB streams.

    Attributes:
        name (str, optional): Name of the stream.
        dim_labels (dict, optional): Dimension labels of the stream.
        file_ext (str, optional): File extension of the stream data.
        is_valid (bool, optional): Indicates if the stream data is valid.
        stream_document_id (ObjectId, optional): ID of the stream document.
        db_sample_rate (float, optional): Sample rate of the stream data in the database.
        type (str, optional): Type of the stream data.
         **kwargs: Arbitrary keyword arguments.
    """

    def __init__(
            self,
            name: str = None,
            dim_labels: dict = None,
            file_ext: str = None,
            is_valid: bool = None,
            stream_document_id: ObjectId = None,
            sr: float = None,
            type: str = None,
            **kwargs,
    ):
        super().__init__(**kwargs)
        self.name = name
        self.dim_labels = dim_labels
        self.file_ext = file_ext
        self.is_valid = is_valid
        self.stream_document_id = stream_document_id
        self.db_sample_rate = sr
        self.type = type


# DATA
class NovaDBHandler:
    """
    Base class for handling MongoDB connections.

     Args:
        db_host (str, optional): IP address of the MongoDB server.
        db_port (int, optional): Port number of the MongoDB server.
        db_user (str, optional): Username for authentication.
        db_password (str, optional): Password for authentication.
        data_dir (str, optional): Datadir to load streams from

    Attributes:
        data_dir (Path, optional): Path to the data directory for stream files
    """

    def __init__(
            self,
            db_host: str = None,
            db_port: int = None,
            db_user: str = None,
            db_password: str = None,
            data_dir: str = None
    ):
        self._client = None
        self._ip = None
        self._port = None
        self._user = None
        self.data_dir = None if data_dir is None else Path(data_dir)
        if db_host and db_port and db_user and db_password:
            self.connect(db_host, db_port, db_user, db_password)

    def connect(
            self, db_host: str = None, db_port: int = None, db_user: str = None, db_password: str = None
    ):
        """
        Connects to the MongoDB server.

        Args:
            db_host (str): IP address or URI of the MongoDB server.
            db_port (int): Port number of the MongoDB server.
            db_user (str): Username for authentication.
            db_password (str): Password for authentication.
        """
        
        # prepend prefix mongodb
        prefix = 'mongodb://'
        if db_host[:len(prefix)] != prefix:
            db_host = prefix+db_host
        
        # Check environment variable for TLS usage
        discover_use_tls = os.getenv('DISCOVER_USE_TLS', '').lower()
        if discover_use_tls in ('true', '1', 'yes'):
            # Use TLS when explicitly enabled
            try:
                self._client = MongoClient(host=db_host, tls=True, tlsAllowInvalidCertificates=True, port=db_port, username=db_user, password=db_password, serverSelectionTimeoutMS=3000)
                self._client.admin.command('ping')  # force lazy connect
            except ServerSelectionTimeoutError:
                print("MongoDB TLS connection attempt failed.")
        elif discover_use_tls in ('false', '0', 'no'):
            # Don't use TLS when explicitly disabled
            try:
                self._client = MongoClient(host=db_host, port=db_port, username=db_user, password=db_password, serverSelectionTimeoutMS=3000)
                self._client.admin.command('ping')  # force lazy connect
            except ServerSelectionTimeoutError:
                print("MongoDB non-TLS connection attempt failed.")
        else:
            # Fall back to original behavior when environment variable is not set or invalid
            try:
                self._client = MongoClient(host=db_host, tls=True, tlsAllowInvalidCertificates=True, port=db_port, username=db_user, password=db_password, serverSelectionTimeoutMS=3000)
                self._client.admin.command('ping')  # force lazy connect
            except ServerSelectionTimeoutError:
                # try without TLS
                try:
                    self._client = MongoClient(host=db_host, port=db_port, username=db_user, password=db_password, serverSelectionTimeoutMS=3000)
                    self._client.admin.command('ping')  # force lazy connect
                except ServerSelectionTimeoutError:
                    print("All MongoDB connection attempts failed.")
           
        self._ip = db_host
        self._port = db_port
        self._user = db_user

    @property
    def client(self) -> MongoClient:
        """
        Returns the MongoDB client instance.
        """
        return self._client


class NovaSession:
    """
    Class to stores all information belonging to a specific session during processing

    Attributes:
        input_data (dict, optional):  Annotation or Stream data that can be processed by a module.
        dataset (str, optional): The dataset or category the session belongs to.
        name (str, optional): The name or title of the session.
        duration (int, optional): The duration of the session in minutes.
        location (str, optional): The location or venue of the session.
        language (str, optional): The language used in the session.
        date (datetime, optional): The date and time of the session.


    Args:
        dataset (str, optional): The dataset or category the session belongs to.
        name (str, optional): The name or title of the session.
        duration (int, optional): The duration of the session in milliseconds.
        location (str, optional): The location or venue of the session.
        language (str, optional): The language used in the session.
        date (datetime, optional): The date and time of the session.
        is_valid (bool, optional): Whether the session is considered valid.
    """

    def __init__(self, input_data: dict = None, dataset: str = None, name: str = None, duration: int = None,
                 location: str = None, language: str = None, date: datetime = None, is_valid: bool = True,
                 extra_data: dict = None, output_data_templates: dict = None):
        self.input_data = input_data
        self.dataset = dataset
        self.name = name
        self.duration = duration
        self.location = location
        self.language = language
        self.date = date


class SessionHandler(NovaDBHandler):
    """
    Handler for loading session data from a MongoDB database.

    This class provides methods to load session information from a MongoDB
    collection and returns a Session object.

    Args:
         (Inherited args from MongoHandler)

    Methods:
        load(dataset: str, session: str) -> Session:
            Load session data from the specified dataset and session name.

    Attributes:
        (Inherited attributes from MongoHandler)

    """

    def load(self, dataset: str, session: Union[str, list, None] = None, keep_order: bool = True) -> list[NovaSession]:
        """
        Load session data from the specified dataset and session name.

        Args:
            dataset (str): The dataset name as specified in the mongo database
            session (str, list, None): The session name as specified in the mongo database. Can be also a list of session names. If no session name is provided, all sessions are loaded instead.
            keep_order (str, bool, True): Keeping sessions in the order of supplied session names.

        Returns:
            NovaSession: A Session object containing loaded session information.
            If the session does not exist, an empty Session object is returned.

        """
        if isinstance(session, str):
            session = [session]

        ret = []
        result = self.client[dataset][SESSION_COLLECTION].find({})

        for s in result:
            if session is not None and s['name'] not in session:
                continue
            # get duration of session in milliseconds
            dur_ms = s.get("duration")
            if dur_ms == 0:
                dur_ms = None
            else:
                dur_ms *= 1000

            ret.append(
                NovaSession(
                    dataset=dataset,
                    name=s["name"],
                    location=s["location"],
                    language=s["language"],
                    date=s["date"],
                    duration=dur_ms,
                    is_valid=s["isValid"],
                )
            )

        # Sorting
        if session is not None and keep_order:
            ret = [r for s in session for r in ret if r.name == s]

        return ret


class AnnotationHandler(IHandler, NovaDBHandler):
    """
    Class for handling download of annotation data from Mongo db.
    """

    def _load_annotation(
            self,
            dataset: str,
            session: str,
            annotator: str,
            role: str,
            scheme: str,
            project: dict = None,
    ) -> dict:
        """
        Load annotation data from MongoDB.

        Args:
            dataset (str): Name of the dataset.
            session (str): Name of the session.
            annotator (str): Name of the annotator.
            role (str): Name of the role.
            scheme (str): Name of the annotation scheme.
            project (dict, optional): Projection for MongoDB query to filter attributes. Defaults to None.

        Returns:
            dict: Loaded annotation data.
        """
        pipeline = [
            {
                "$lookup": {
                    "from": SESSION_COLLECTION,
                    "localField": "session_id",
                    "foreignField": "_id",
                    "as": "session",
                }
            },
            {
                "$lookup": {
                    "from": ANNOTATOR_COLLECTION,
                    "localField": "annotator_id",
                    "foreignField": "_id",
                    "as": "annotator",
                }
            },
            {
                "$lookup": {
                    "from": ROLE_COLLECTION,
                    "localField": "role_id",
                    "foreignField": "_id",
                    "as": "role",
                }
            },
            {
                "$lookup": {
                    "from": SCHEME_COLLECTION,
                    "localField": "scheme_id",
                    "foreignField": "_id",
                    "as": "scheme",
                }
            },
            {
                "$match": {
                    "$and": [
                        {"role.name": role},
                        {"session.name": session},
                        {"annotator.name": annotator},
                        {"scheme.name": scheme},
                    ]
                }
            },
            {
                "$lookup": {
                    "from": "AnnotationData",
                    "localField": "data_id",
                    "foreignField": "_id",
                    "as": "data",
                }
            },
        ]

        # append projection
        if project:
            pipeline.append({"$project": project})

        result = list(self.client[dataset][ANNOTATION_COLLECTION].aggregate(pipeline))
        if not result:
            return {}
        return result[0]

    def _load_scheme(self, dataset: str, scheme: str) -> dict:
        result = self.client[dataset][SCHEME_COLLECTION].find_one({"name": scheme})
        if not result:
            return {}
        return result

    def _update_annotation(
            self,
            dataset: str,
            annotation_id: ObjectId,
            annotation_data_id: ObjectId,
            annotation_data: list[dict],
            is_finished: bool,
            is_locked: bool,
    ) -> UpdateResult:
        """
        Updates existing annotation the Mongo database

        Args:
            dataset (str): Name of the dataset.
            annotation_id (ObjectId): ObjectId of the annotation in the database
            annotation_data_id (ObjectId): ObjectId of the corresponding annotation data object in the database
            annotation_data (list[dict]): List of dictionaries containing the annotation data. Each dictionary represents one sample. Keys must match the annotation types.
            is_finished (bool): Whether the annotation has already been fully completed or not
            is_locked (bool): Whether the annotation should be locked and can therefore not be overwritten anymore.

        Returns:
            UpdateResult: The success status of the update operation
        """
        update_query_annotation = {
            "$set": {
                "date": datetime.now(),
                "isFinished": is_finished,
                "isLocked": is_locked,
            }
        }
        update_query_annotation_data = {"$set": {"labels": annotation_data}}
        success = self.client[dataset][ANNOTATION_COLLECTION].update_one(
            {"_id": annotation_id}, update_query_annotation
        )

        if not success.acknowledged:
            return success

        success = self.client[dataset][ANNOTATION_DATA_COLLECTION].update_one(
            {"_id": annotation_data_id}, update_query_annotation_data
        )

        return success

    def _insert_annotation_data(self, dataset: str, data: list) -> InsertOneResult:
        """
        Insert annotation data into the MongoDB database.

        Args:
            dataset (str): Name of the dataset.
            data (list): List of annotation data to be inserted.

        Returns:
            InsertOneResult: The result of the insertion operation.
        """
        annotation_data = {"labels": data}

        success = self.client[dataset][ANNOTATION_DATA_COLLECTION].insert_one(
            annotation_data
        )
        return success

    def _insert_annotation(
            self,
            dataset: str,
            session_id: ObjectId,
            annotator_id: ObjectId,
            scheme_id: ObjectId,
            role_id: ObjectId,
            data: list,
            is_finished: bool,
            is_locked: bool,
    ):
        """
        Insert annotation and associated annotation data into the MongoDB database.

        Args:
            dataset (str): Name of the dataset.
            session_id (ObjectId): ID of the associated session.
            annotator_id (ObjectId): ID of the annotator.
            scheme_id (ObjectId): ID of the annotation scheme.
            role_id (ObjectId): ID of the role.
            data (list): List of annotation data to be inserted.
            is_finished (bool): Indicates if the annotation is finished.
            is_locked (bool): Indicates if the annotation is locked.

        Returns:
            InsertOneResult: The result of the insertion operation.
        """
        # insert annotation data first
        success = self._insert_annotation_data(dataset, data)
        if not success.acknowledged:
            return success
        else:
            data_id = success.inserted_id

        # insert annotation object
        annotation_document = {
            "session_id": session_id,
            "annotator_id": annotator_id,
            "scheme_id": scheme_id,
            "role_id": role_id,
            "date": datetime.now(),
            "isFinished": is_finished,
            "isLocked": is_locked,
            "data_id": data_id,
        }
        success = self.client[dataset][ANNOTATION_COLLECTION].insert_one(
            annotation_document
        )

        # if the annotation could not be created we delete the annotation data as well
        if not success.acknowledged:
            success = self.client[dataset][ANNOTATION_DATA_COLLECTION].delete_one(
                {"_id": data_id}
            )

        return success

    def load(
            self, dataset: str, scheme: str, session: str, annotator: str, role: str, header_only: bool = False
    ) -> Annotation:
        """
        Load annotation data from MongoDB and create an Annotation object.

        Args:
            dataset (str): Name of the dataset.
            scheme (str): Name of the annotation scheme.
            session (str): Name of the session.
            annotator (str): Name of the annotator.
            role (str): Name of the role.
            header_only (bool): If true only the annotation header will be loaded.


        Returns:
            Annotation: An Annotation object loaded from the database.

        Raises:
            FileNotFoundError: If the requested annotation data is not found in the database.
            TypeError: If the scheme type is unknown.
        """
        # load annotation from mongo db
        if header_only:
            scheme_doc = self._load_scheme(dataset, scheme)

            if not scheme_doc:
                raise FileNotFoundError(
                    f"Scheme not found dataset: {dataset} scheme: {scheme}"
                )

        else:
            anno_doc = self._load_annotation(dataset, session, annotator, role, scheme)

            if not anno_doc:
                raise FileNotFoundError(
                    f"Annotation not found dataset: {dataset} session: {session} annotator: {annotator} role: {role} scheme: {scheme}"
                )

            (anno_data_doc,) = anno_doc["data"]

            # build annotation object
            (scheme_doc,) = anno_doc["scheme"]

        scheme_type = scheme_doc["type"]
        scheme_description = scheme_doc.get("description", '')
        scheme_examples = scheme_doc.get("examples", [])
        scheme_attributes = scheme_doc.get("attributes", [])
        for sa in scheme_attributes:
            sa['values'] = [str(v['value']) for v in sa['values']]

        anno_data = None
        anno_duration = 0
        meta_data = None

        # discrete scheme
        if scheme_type == SchemeType.DISCRETE.name:
            #scheme_classes = {l["id"]: l["name"] for l in scheme_doc["labels"]}
            scheme_classes = {l["id"]: l for l in scheme_doc["labels"]}
            if not header_only:
                    anno_data = []
                    meta_data = []
                    for x in anno_data_doc['labels']:
                        anno_data.append((x["from"], x["to"], x["id"], x["conf"]))
                        attribute = x.get('meta', '')  # empty dict instead of empty string
                        if attribute == '':
                            attribute = {}
                        else:
                            attribute = attribute[len('attributes:'):]  # parse attributes string
                            if not scheme_attributes:
                                raise ValueError(f"Annotation has attribute values '{attribute}' but scheme '{scheme}' has no defined attributes. Please update the annotation scheme to include attribute definitions.")
                            # Fix malformed dict/json by quoting keys and values
                            import re
                            import json

                            def escape_and_quote(match):
                                key = match.group(1)
                                value = match.group(2)
                                # Use json.dumps to properly escape the value
                                escaped_value = json.dumps(value)
                                return f'"{key}":{escaped_value}'

                            # Quote key:{value} -> "key":"properly_escaped_value"
                            attribute = re.sub(r'([^:,{}]+):\{([^}]+)\}', escape_and_quote, attribute)
                            attribute = json.loads(attribute)  # Use json.loads instead of eval
                        meta_data.append(attribute)

                    # Convert row-wise attribute format to column-wise format for compatibility
                    if meta_data and scheme_attributes:
                        # Collect all unique attribute keys from schema
                        all_keys = {attr['name'] for attr in scheme_attributes}

                        # Convert from row-wise list of dicts to column-wise dict of lists
                        column_wise_attributes = {}
                        for key in all_keys:
                            column_wise_attributes[key] = []
                            for row_attributes in meta_data:
                                # Use None as default for missing attributes in empty rows
                                column_wise_attributes[key].append(row_attributes.get(key, None))

                        meta_data = column_wise_attributes
                    elif meta_data:
                        # If we have meta_data but no scheme_attributes, keep original format for now
                        # This maintains backward compatibility
                        pass
                    else:
                        # No attributes at all
                        meta_data = None

                    anno_data = np.array(anno_data, dtype=SSILabelDType.DISCRETE.value)
                    anno_data = convert_ssi_to_label_dtype(anno_data, SchemeType.DISCRETE)
                    anno_duration = anno_data[-1]["to"] if anno_data.size != 0 else 0

            anno_scheme = DiscreteAnnotationScheme(name=scheme, classes=scheme_classes)
            annotation = DiscreteAnnotation(
                role=role,
                session=session,
                dataset=dataset,
                data=anno_data,
                scheme=anno_scheme,
                annotator=annotator,
                duration=anno_duration
            )
            annotation.meta_data.description = scheme_description
            annotation.meta_data.examples = scheme_examples
            if scheme_attributes:
                annotation.meta_data.attributes = scheme_attributes
                annotation.meta_data.attribute_values = meta_data

        # continuous scheme
        elif scheme_type == SchemeType.CONTINUOUS.name:
            sr = scheme_doc["sr"]
            min_val = scheme_doc["min"]
            max_val = scheme_doc["max"]

            if not header_only:
                anno_data = np.array(
                    [(x["score"], x["conf"]) for x in anno_data_doc["labels"]],
                    dtype=SSILabelDType.CONTINUOUS.value,
                )
                anno_data = convert_ssi_to_label_dtype(anno_data, SchemeType.CONTINUOUS)
                anno_duration = len(anno_data_doc["labels"]) / sr

            anno_scheme = ContinuousAnnotationScheme(
                name=scheme, sample_rate=sr, min_val=min_val, max_val=max_val
            )
            annotation = ContinuousAnnotation(
                role=role,
                session=session,
                dataset=dataset,
                data=anno_data,
                scheme=anno_scheme,
                annotator=annotator,
                duration=anno_duration,
            )

            # free scheme

        # free scheme
        elif scheme_type == SchemeType.FREE.name:

            if not header_only:
                anno_data = np.array(
                    [
                        (x["from"], x["to"], x["name"], x["conf"])
                        for x in anno_data_doc["labels"]
                    ],
                    dtype=SSILabelDType.FREE.value,
                )

                anno_data = convert_ssi_to_label_dtype(anno_data, SchemeType.FREE)
                anno_duration = anno_data[-1]["to"] if anno_data.size != 0 else 0

            anno_scheme = FreeAnnotationScheme(name=scheme)
            annotation = FreeAnnotation(
                role=role,
                session=session,
                dataset=dataset,
                data=anno_data,
                scheme=anno_scheme,
                annotator=annotator,
                duration=anno_duration,
            )
        else:
            raise TypeError(f"Unknown scheme type {scheme_type}")

        annotation.meta_data.examples = scheme_examples
        annotation.meta_data.description = scheme_description
        annotation.meta_data.attributes = scheme_attributes

        # setting meta data
        if header_only:
            handler_meta_data = NovaDBAnnotationMetaData(
                ip=self._ip,
                port=self._port,
                user=self._user,
                dataset=dataset
            )
        else:
            handler_meta_data = NovaDBAnnotationMetaData(
                ip=self._ip,
                port=self._port,
                user=self._user,
                dataset=dataset,
                is_locked=anno_doc.get("isLocked"),
                is_finished=anno_doc.get("isFinished"),
                annotation_document_id=anno_doc.get("_id"),
                data_document_id=anno_doc.get("data_id"),
                last_update=anno_doc.get("date"),
            )
        annotation.meta_data.expand(handler_meta_data)

        return annotation

    def save(
            self,
            annotation: Annotation,
            dataset: str = None,
            session: str = None,
            annotator: str = None,
            role: str = None,
            scheme: str = None,
            is_finished: bool = False,
            is_locked: bool = False,
            overwrite: bool = False,
    ):
        """
        Save an Annotation object to the MongoDB database.

        Args:
            annotation (Annotation): The Annotation object to be saved.
            dataset (str, optional): Name of the dataset. Overwrites the respective attribute from annotation.meta_data if set. Defaults to None.
            session (str, optional): Name of the session. Overwrites the respective attribute from annotation.meta_data if set. Defaults to None.
            annotator (str, optional): Name of the annotator. Overwrites the respective attribute from annotation.meta_data if set. Defaults to None.
            role (str, optional): Name of the role. Overwrites the respective attribute from annotation.meta_data if set. Defaults to None.
            scheme (str, optional): Name of the annotation scheme. Overwrites the respective attribute from annotation.meta_data if set. Defaults to None.
            is_finished (bool, optional): Indicates if the annotation is finished. Defaults to False.
            is_locked (bool, optional): Indicates if the annotation is locked. Defaults to False.
            overwrite (bool, optional): If True, overwrite an existing annotation. Defaults to False.

        Returns:
            UpdateResult: The result of the update operation.

        Raises:
            FileExistError: If annotation exists and is locked or annotation exists and overwrite is set to false
        """
        # overwrite default values
        dataset = dataset if not dataset is None else annotation.meta_data.dataset
        session = session if not session is None else annotation.meta_data.session
        annotator = annotator if not annotator is None else annotation.meta_data.annotator
        role = role if not role is None else annotation.meta_data.role
        scheme = scheme if not scheme is None else annotation.annotation_scheme.name

        if isinstance(annotation.data, np.ndarray) and not annotation.data.size == 0:

            # Assure samplerates for continuous annotations are matching
            if isinstance(annotation, ContinuousAnnotation):
                scheme_doc = self._load_scheme(dataset, scheme)
                sr = scheme_doc['sr']
                annotation.data = resample(annotation.data, src_sr=annotation.annotation_scheme.sample_rate, trgt_sr=sr)

            if isinstance(annotation, DiscreteAnnotation):
                # Remove rest class
                non_rest_class_idxs = (annotation.data['id'] != annotation.rest_label_id)
                annotation.data = annotation.data[non_rest_class_idxs]
                if annotation.meta_data.attribute_values is not None:
                    for k in annotation.meta_data.attribute_values.keys():
                       annotation.meta_data.attribute_values[k] = list(np.asarray(annotation.meta_data.attribute_values[k])[non_rest_class_idxs])

            anno_data = convert_label_to_ssi_dtype(
                annotation.data, annotation.annotation_scheme.scheme_type
            )

            # TODO check for none values
            anno_data = [
                dict(zip(annotation.annotation_scheme.label_dtype.names, ad.item()))
                for ad in anno_data
            ]

            if annotation.meta_data.attribute_values is not None:
                for k in list(annotation.meta_data.attribute_values.keys()):
                    if len(annotation.meta_data.attribute_values[k]) != len(anno_data):
                        annotation.meta_data.attribute_values.pop(k)
                        warnings.warn(
                            f"Number of values for attribute '{k}' do not match number of samples. Attribute will not be saved to the database"
                        )
                for i, ad in enumerate(anno_data):
                    ad['meta'] = "attributes:{" + ','.join(
                        [(str(k) + ":{" + str(v[i]) + "}") for k, v in annotation.meta_data.attribute_values.items()]) + "}"

        else:
            anno_data = []

        # load annotation to check if an annotation for the provided criteria already exists in the database
        anno_doc = self._load_annotation(
            dataset,
            session,
            annotator,
            role,
            scheme,
            project={"_id": 1, "isLocked": 1, "data_id": 1},
        )

        # update existing annotations
        if anno_doc:
            if anno_doc["isLocked"]:
                raise FileExistsError(
                    f"Can't overwrite locked annotation dataset: {dataset} session: {session} annotator: {annotator} role: {role} scheme: {scheme}. Because annotation is locked."
                )
            elif not overwrite:
                raise FileExistsError(
                    f"Can't overwrite annotation dataset: {dataset} session: {session} annotator: {annotator} role: {role} scheme: {scheme}. Because overwrite is disabled."
                )
            else:
                warnings.warn(
                    f"Overwriting existing annotation dataset: {dataset} session: {session} annotator: {annotator} role: {role} scheme: {scheme}"
                )

                success = self._update_annotation(
                    dataset=dataset,
                    annotation_id=anno_doc["_id"],
                    annotation_data_id=anno_doc["data_id"],
                    annotation_data=anno_data,
                    is_finished=is_finished,
                    is_locked=is_locked,
                )

        # add new annotation
        else:
            scheme_id = self.client[dataset][SCHEME_COLLECTION].find_one(
                {"name": scheme}
            )["_id"]
            session_id = self.client[dataset][SESSION_COLLECTION].find_one(
                {"name": session}
            )["_id"]
            role_id = self.client[dataset][ROLE_COLLECTION].find_one({"name": role})[
                "_id"
            ]
            annotator_id = self.client[dataset][ANNOTATOR_COLLECTION].find_one(
                {"name": annotator}
            )["_id"]
            success = self._insert_annotation(
                dataset=dataset,
                scheme_id=scheme_id,
                session_id=session_id,
                annotator_id=annotator_id,
                role_id=role_id,
                data=anno_data,
                is_finished=is_finished,
                is_locked=is_locked,
            )

        return success
        # TODO success error handling


class StreamHandler(IHandler, NovaDBHandler):
    """
    Class for handling download and upload of stream data from MongoDB.
    """

    def _load_stream(
            self,
            dataset: str,
            stream_name: str,
    ) -> dict:
        """
        Load stream data from MongoDB.

        Args:
            dataset (str): Name of the dataset.
            stream_name (str): Name of the stream.

        Returns:
            dict: Loaded stream data.
        """
        result = self.client[dataset][STREAM_COLLECTION].find_one({"name": stream_name})
        if not result:
            return {}
        return result

    def load(self, dataset: str, session: str, role: str, name: str, header_only: bool = False) -> Stream:
        """
        Load a Stream object from MongoDB and create a Stream instance.

        Args:
            dataset (str): Name of the dataset.
            session (str): Name of the session.
            role (str): Name of the role.
            name (str): Name of the stream.
            header_only (bool): If true only the stream header will be loaded.

        Returns:
            Stream: A Stream object loaded from the database.

        Raises:
            ValueError: If the requested stream is not found for the given dataset.
            FileNotFoundError: If the data directory is not set or the file is not found on disc.
        """
        result = self._load_stream(dataset=dataset, stream_name=name)
        if not result:
            raise FileNotFoundError(f"No stream {name} found for dataset {dataset}")
        if not self.data_dir:
            raise FileNotFoundError("Data directory was not set. Can't access files")

        file_path = Path(
            self.data_dir
            / dataset
            / session
            / (role + "." + result["name"] + "." + result["fileExt"])
        )

        if not file_path.is_file():
            raise FileNotFoundError(f"No such file {file_path}")

        # data
        data = FileHandler().load(file_path, header_only=header_only)
        assert isinstance(data, Stream)

        # meta data
        data.meta_data.role = role
        data.meta_data.name = name
        data.meta_data.session = session

        handler_meta_data = NovaDBStreamMetaData(
            ip=self._ip,
            port=self._port,
            user=self._user,
            name=result.get("name"),
            dim_labels=result.get("dimlabels"),
            file_ext=result.get("fileExt"),
            is_valid=result.get("isValid"),
            stream_document_id=result.get("_id"),
            sr=result.get("sr"),
            type=result.get("type"),
            dataset=dataset
        )
        data.meta_data.expand(handler_meta_data)

        return data

    def save(
            self,
            stream: Stream,
            dataset: str = None,
            session: str = None,
            role: str = None,
            name: str = None,
            media_type: str = None,
            file_ext: str = None,
            dim_labels: list = None,
            is_valid: bool = True,
    ):
        """
        Save a Stream object to the MongoDB database and store associated file.

        Args:
            stream (Stream): The Stream object to be saved.
            dataset (str): Name of the dataset. Overwrites the respective attribute from stream.meta_data if set. Defaults to None.
            session (str): Name of the session. Overwrites the respective attribute from stream.meta_data if set. Defaults to None.
            role (str): Name of the role. Overwrites the respective attribute from stream.meta_data if set. Defaults to None.
            name (str): Name of the stream. Overwrites the respective attribute from stream.meta_data if set. Defaults to None.
            media_type (str): Media type of the stream data as specified in NOVA-DB.
            file_ext (str, optional): File extension. Defaults to None.
            dim_labels (list[dict], optional): Dimension labels. Defaults to None.
            is_valid (bool, optional): Indicates if the stream data is valid. Defaults to True.

        Raises:
            FileNotFoundError: If the data directory is not set.
        """

        # save file to disk
        dataset = dataset if not dataset is None else stream.meta_data.dataset
        session = session if not session is None else stream.meta_data.session
        role = role if not role is None else stream.meta_data.role
        name = name if not name is None else stream.meta_data.name
        file_ext = file_ext if not file_ext is None else stream.meta_data.ext
        media_type = media_type if not media_type is None else stream.meta_data.media_type

        if isinstance(stream, SSIStream):
            dim_labels = dim_labels if not dim_labels is None else stream.meta_data.dim_labels
        else:
            dim_labels = None

        if not self.data_dir:
            raise FileNotFoundError("Data directory was not set. Can't access files")

        if not self.data_dir.is_dir():
            raise NotADirectoryError(f"Specified data directory {self.data_dir} is not a directory.")

        file_name = role + "." + name + file_ext
        file_path = self.data_dir / dataset / session / file_name

        FileHandler().save(stream, file_path)

        meta_data: StreamMetaData = stream.meta_data

        # write db entry
        # NOTE: "dimlabels" is lowercase to match NOVA's DatabaseHandler.cs convention.
        # Other fields use camelCase (fileExt, isValid) — this inconsistency originates in NOVA.
        # If NOVA compatibility is ever dropped, consider normalizing all field names, eg snake_case.
        stream_document = {
            "fileExt": file_ext.strip('.'),
            "name": name,
            "sr": meta_data.sample_rate,
            "type": media_type,
            "dimlabels": dim_labels if dim_labels else [],
            "isValid": is_valid,
        }

        # check if stream exists
        result = self.client[dataset][STREAM_COLLECTION].find_one({"name": name})

        # update existing
        if result:
            update_query_annotation = {"$set": stream_document}
            self.client[dataset][STREAM_COLLECTION].update_one(
                {"_id": result["_id"]}, update_query_annotation
            )

        # insert new
        else:
            self.client[dataset][STREAM_COLLECTION].insert_one(stream_document)


if __name__ == "__main__":
    import os
    import random
    from time import perf_counter
    import dotenv

    test_annotations = True
    test_streams = True

    dotenv.load_dotenv()
    IP = os.getenv("NOVA_IP", "")
    PORT = int(os.getenv("NOVA_PORT", 0))
    USER = os.getenv("NOVA_USER", "")
    PASSWORD = os.getenv("NOVA_PASSWORD", "")
    DATA_DIR = os.getenv("NOVA_DATA_DIR", None)


    DATASET = os.getenv("DISCOVER_ITERATOR_TEST_DATASET")
    SESSION = os.getenv("DISCOVER_ITERATOR_TEST_SESSION")
    SCHEME = os.getenv("DISCOVER_ITERATOR_TEST_SCHEME")
    ANNOTATOR = os.getenv("DISCOVER_ITERATOR_TEST_ANNOTATOR")
    ROLE = os.getenv("DISCOVER_ITERATOR_TEST_ROLE")
    FEATURE_STREAM = os.getenv("DISCOVER_ITERATOR_TEST_STREAM")

    if test_annotations:
        amh = AnnotationHandler(db_host=IP, db_port=PORT, db_user=USER, db_password=PASSWORD)

        # load
        fs = "Loading {} took {}ms"
        t_start = perf_counter()
        anno = amh.load(
            dataset=DATASET,
            scheme=SCHEME,
            annotator=ANNOTATOR,
            session=SESSION,
            role=ROLE,
            header_only=False
        )
        t_stop = perf_counter()
        print(fs.format("Loaded annotation", int((t_stop - t_start) * 1000)))

        # save
        fs = "Saving {} took {}ms"
        t_start = perf_counter()
        amh.save(
            dataset=DATASET,
            annotation=anno,
            session=SESSION,
            role=ROLE,
            overwrite=True,
        )
        t_stop = perf_counter()
        print(fs.format("Discrete annotation", int((t_stop - t_start) * 1000)))


    if test_streams:
        smh = StreamHandler(
            db_host=IP, db_port=PORT, db_user=USER, db_password=PASSWORD, data_dir=Path(DATA_DIR)
        )

        # Stream
        fs = "Loading {} took {}ms"
        t_start = perf_counter()
        feature_stream = smh.load(
            dataset=DATASET,
            session=SESSION,
            role=ROLE,
            name=FEATURE_STREAM,
            header_only=False
        )
        t_stop = perf_counter()
        print(fs.format("Video", int((t_stop - t_start) * 1000)))

        suffix = "_testing"
        feature_stream.sample_rate = random.uniform(0, 16000)
        smh.save(
            stream=feature_stream,
            dataset=DATASET,
            session=SESSION,
            role=ROLE,
            name=FEATURE_STREAM + suffix,
            media_type="feature",
            dim_labels=[{"id": 1, "name": "hallo"}, {"id": 2, "name": "nope"}],
        )

        # Audio
        t_start = perf_counter()
        audio_stream = smh.load(
            dataset=DATASET, session=SESSION, role=ROLE, name="audio"
        )
        t_stop = perf_counter()
        print(fs.format("Audio", int((t_stop - t_start) * 1000)))

        # Video
        t_start = perf_counter()
        video_stream = smh.load(
            dataset=DATASET, session=SESSION, role=ROLE, name="video"
        )
        t_stop = perf_counter()
        print(fs.format("Video", int((t_stop - t_start) * 1000)))

        breakpoint()



""" Module to create a data iterator for streams and annotations

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    18.8.2023

"""
import os
import sys
import warnings
from pathlib import Path
from typing import Union
from collections.abc import Iterator

import numpy as np
from discover_utils.data.handler.nova_db_handler import NovaSession
from discover_utils.data.provider.data_manager import SessionManager, DatasetManager
from discover_utils.data.stream import Stream, Audio
from discover_utils.data.annotation import FreeAnnotation, DiscreteAnnotation, ContinuousAnnotation
from discover_utils.utils import string_utils
from discover_utils.utils.anno_utils import data_contains_garbage


class DatasetIterator(DatasetManager):
    """Iterator class for processing data samples from the Nova dataset.

    The NovaIterator takes all information about what data should be loaded and how it should be processed. The class itself then takes care of loading all data and provides an iterator to directly apply a sliding window to the requested data.
    Every time based argument can be passed either as string or a numerical value. If the time is passed as string, the string should end with either 's' to indicate the time is specified in seconds or 'ms' for milliseconds.
    If the time is passed as a numerical value or as a string without indicating a specific unit it is assumed that an integer value represents milliseconds while a float represents seconds. All numbers will be represented as integer milliseconds internally.
    The highest time resolution for processing is therefore 1ms.

    Args:


        frame_size (Union[int, float, str], optional): Size of the data frame measured in time. Defaults to None.
        start (Union[int, float, str], optional): Start time for processing measured in time. Defaults to None.
        end (Union[int, float, str], optional): End time for processing measured in time. Defaults to None.
        left_context (Union[int, float, str], optional): Left context duration measured in time. Defaults to None.
        right_context (Union[int, float, str], optional): Right context duration measured in time. Defaults to None.
        stride (Union[int, float, str], optional): Stride for iterating over data measured in time. If stride is not set explicitly it will be set to frame_size. Defaults to None.
        add_rest_class (bool, optional): Whether to add a rest class for discrete annotations. Defaults to True.
        fill_missing_data (bool, optional): Whether to fill missing data. Defaults to True. THIS OPTION IS CURRENTLY NOT DOING ANYTHING

    Attributes:


        frame_size (int, float, str): Size of the data frame measured in time.
        start (int, float, str): Start time for processing measured in time.
        end (int, float, str): End time for processing measured in time.
        left_context (int, float, str): Left context duration measured in time.
        right_context (int, float, str): Right context duration measured in time.
        stride (int, float, str): Stride for iterating  data measured in time.
        add_rest_class (bool): Whether to add a rest class for discrete annotations.
        fill_missing_data (bool): Whether to fill missing data.

    Example:
        .. code-block:: python

            from dotenv import load_dotenv

            # Load environment variables
            IP = 127.0.0.1
            PORT = 1337
            USER = my_user
            PASSWORD = my_password
            DATA_DIR = /data

            # Define dataset and sessions
            dataset = "test_dataset"
            sessions = ["test_session"]

            # Define data descriptions
            annotation = {
                "id" : "my_transcript_1",
                "type": "input",
                "src": "db:anno",
                "scheme": "transcript",
                "annotator": "test_annotator",
                "role": "test_role",
            }

            stream = {
                "id" : "model_output",
                "type" : "input",
                "src": "db:stream",
                "role": "test_role",
                "name": "extracted_features",
            }

            file = {
                "id": "just_a_file",
                "type": "output",
                "src": "file:stream",
                "fp": "/path/to/my/video/test_video.mp4",
            }

            # Create a NovaIterator instance
            nova_iterator = NovaIterator(
                IP,
                PORT,
                USER,
                PASSWORD,
                dataset,
                DATA_DIR,
                sessions=sessions,
                data=[annotation, file],
                frame_size="5s",
                end="20s",
            )

            # Example: Get the next data sample and set a breakpoint
            a = next(nova_iterator)
    """

    def __init__(
            self,
            # Data
            #dataset_manager: DatasetManager,
            *args,

            # Iterator Window
            frame_size: Union[int, float, str] = None,
            start: Union[int, float, str] = None,
            end: Union[int, float, str] = None,
            left_context: Union[int, float, str] = None,
            right_context: Union[int, float, str] = None,
            stride: Union[int, float, str] = None,

            # Iterator properties
            add_rest_class: bool = True,
            fill_missing_data=True,

            **kwargs
    ):
        #self.dataset_manager = dataset_manager
        super().__init__(*args, **kwargs)
        # If stride has not been explicitly set it's the same as the frame size
        if stride is None:
            stride = frame_size

        # Parse all times to milliseconds
        self.left_context = string_utils.parse_time_string_to_ms(left_context)
        self.right_context = string_utils.parse_time_string_to_ms(right_context)
        self.frame_size = string_utils.parse_time_string_to_ms(frame_size)
        self.stride = string_utils.parse_time_string_to_ms(stride)
        self.start = string_utils.parse_time_string_to_ms(start)
        self.end = string_utils.parse_time_string_to_ms(end)

        # Frame size 0 or None indicates that the whole session should be returned as one sample
        if self.frame_size == 0:
            warnings.warn(
                "Frame size should be bigger than zero. Returning whole session as sample."
            )

        # If the end time has not been set we initialize it with sys.maxsize
        if self.end is None or self.end == 0:
            self.end = sys.maxsize

        self.add_rest_class = add_rest_class
        self.fill_missing_data = fill_missing_data
        self.current_session = None

        # Data handler
        self._iterable = self._yield_sample()

    def _guess_session_length(self, session):

        annotation_duration = -1
        stream_duration = sys.maxsize
        for id, data in session.input_data.items():
            # Discrete and free annotations do not need to be present at the end of a session. So we consider the maximum length.
            if isinstance(data, FreeAnnotation) or isinstance(data, DiscreteAnnotation):
                if data.meta_data.duration is not None and data.meta_data.duration > annotation_duration:
                    annotation_duration = data.meta_data.duration
            # Streams and continuous annotations do need to cover the whole session. Consider the minimum length.
            elif isinstance(data, Stream) or isinstance(data, ContinuousAnnotation):
                if data.meta_data.duration is not None and data.meta_data.duration < stream_duration:
                    stream_duration = data.meta_data.duration

        # If a datastream is present we use it to determine the length
        if stream_duration < sys.maxsize:
            return stream_duration
        else:
            return annotation_duration

    def _yield_sample(self) -> Iterator[dict]:
        """
        Yield examples.

        Yields:
            dict: Sampled data for the window.

        Example:
            {
                <scheme>_<annotator>_<role> : [[data1], [data2]... ],
                <file_path> : [[data1], [data2]... ]
            }
        """

        # Needed to sort the samples later and assure that the order is the same as in nova.
        # sample_counter = 1

        for session_name, session in self.sessions.items():
            # Init all data objects for the session and get necessary meta information
            self.session_manager: SessionManager
            self.session_info: NovaSession
            self.current_session = session.get('manager')
            self.current_session_info = session.get('info')
            self.current_session.load()

            # TODO: Check if we run into duration errors
            if self.current_session_info is None:
                self.current_session_info = NovaSession()
            if self.current_session_info.duration is None:
                self.current_session_info.duration = self._guess_session_length(self.current_session)

            # If frame size is zero or less we return the whole data from the whole session in one sample
            if self.frame_size <= 0:
                _frame_size = min(self.current_session_info.duration, self.end - self.start)
                _stride = _frame_size
            else:
                _frame_size = self.frame_size
                _stride = self.stride

            # Starting position of the first frame in seconds
            cpos = 0

            # TODO account for stride and framesize being None
            # Generate samples for this session
            while cpos + self.stride <= min(
                    self.end, self.current_session_info.duration
            ):
                frame_start = cpos
                frame_end = cpos + _frame_size

                window_start = frame_start - self.left_context
                window_end = frame_end + self.right_context

                window_info = (
                        self.current_session.session
                        + "_"
                        + str(window_start / 1000)
                        + "_"
                        + str(window_end / 1000)
                )

                data_for_window = {}
                warn_once = []

                for k, v in self.current_session.input_data.items():

                    # TODO current_session duration is not the correct way to end right padding. We could have longer streams
                    start_ = max(0, window_start)
                    end_ = min(self.current_session_info.duration, window_end)

                    try:
                        sample = v.sample_from_interval(start_, end_)
                    except Exception as e:
                        # Stream is shorter than session duration
                        if v.meta_data.__dict__.get('duration', sys.maxsize) < self.current_session_info.duration:
                            if not k in warn_once:
                                warn_once.append(k)
                                print(f'Error retreiving data for window {window_info} in stream {k}. Stream is shorter than session. From now on np.NaN values will be returned')

                        else:
                            print(f'Error retreiving data for window {window_info} in stream {k}. Exception caught when retreiving sample. Returning NaN values: {e}')
                        data_for_window[k] = np.nan

                    # TODO pad continuous annotations
                    # Apply padding
                    if isinstance(v, Stream):

                        # Don't pad anything but num_samples axis
                        sr = v.meta_data.sample_rate / 1000
                        left_pad = int((0 - window_start) * sr) if window_start < 0 else 0
                        right_pad = int((
                                                    window_end - self.current_session_info.duration) * sr) if window_end > self.current_session_info.duration else 0

                        # Currently all streams have timedim first. Keep for future reference.
                        time_dim_last = False

                        # In some cases sample sample_from_interval might return a frame number that is one frame off from what we expect.
                        # This is due to sampling issues when frame sizes do not match the samplerate. We fix this here.
                        num_samples_exp = int((abs(window_start) + window_end) * sr)
                        num_samples = int(left_pad + right_pad + sample.shape[0])
                        if num_samples > num_samples_exp:
                            diff = (num_samples - num_samples_exp)
                            if time_dim_last:
                                sample = sample[:,:-diff]
                            else:
                                sample = sample[:-diff]


                        if left_pad or right_pad:
                            lr_pad = ((left_pad, right_pad),)
                            n_pad = tuple([(0, 0)] * (len(sample.shape) - 1))

                            # Num samples last dim
                            if time_dim_last:
                                pad = n_pad + lr_pad
                            # Num samples first dim
                            else:
                                pad = lr_pad + n_pad

                            sample = np.pad(sample, pad_width=pad, mode='edge')
                    data_for_window[k] = sample

                # Performing sanity checks
                garbage_detected = any(
                    [data_contains_garbage(d) for k, d in data_for_window.items()]
                )

                # Incrementing counter
                cpos += _stride

                if garbage_detected:
                    continue

                data_for_window['info'] = window_info
                yield data_for_window

    def __iter__(self) -> Iterator[dict]:
        return self._iterable

    def __next__(self) -> dict:
        return self._iterable.__next__()


if __name__ == "__main__":
    from pathlib import Path
    import os
    import dotenv
    dotenv.load_dotenv()
    DISCOVER_TEST_FILE_DIR = Path(os.getenv("DISCOVER_DATA_DIR"))
    DISCOVER_OUR_DIR = Path(os.getenv("DISCOVER_TEST_DIR"))

    IP = os.getenv("NOVA_IP", "")
    PORT = int(os.getenv("NOVA_PORT", 0))
    USER = os.getenv("NOVA_USER", "")
    PASSWORD = os.getenv("NOVA_PASSWORD", "")
    DATA_DIR = Path(os.getenv("NOVA_DATA_DIR", None))

    DATASET = os.getenv("DISCOVER_ITERATOR_TEST_DATASET")
    SESSIONS = [os.getenv("DISCOVER_ITERATOR_TEST_SESSION")]
    SCHEME = os.getenv("DISCOVER_ITERATOR_TEST_SCHEME")
    ANNOTATOR = os.getenv("DISCOVER_ITERATOR_TEST_ANNOTATOR")
    ROLE = os.getenv("DISCOVER_ITERATOR_TEST_ROLE")
    FEATURE_STREAM = os.getenv("DISCOVER_ITERATOR_TEST_STREAM")

    annotation = {
        "src": "db:annotation",
        "scheme": SCHEME,
        "type": "input",
        "id": "annotation",
        "annotator": ANNOTATOR,
        "role": ROLE,
    }

    stream = {
        "src": "db:stream",
        "type": "input",
        "id": "featurestream",
        "role": ROLE,
        "name": FEATURE_STREAM,
    }

    file = {
        "src": "file:stream:Audio",
        "type": "input",
        "id": "file",
        "uri": DISCOVER_TEST_FILE_DIR/"test_audio.wav",
    }

    ctx = {
        "db": {
            "db_host": IP,
            "db_port": PORT,
            "db_user": USER,
            "db_password": PASSWORD,
            "data_dir": DATA_DIR,
        },
    }

    nova_iterator = DatasetIterator(
        dataset=DATASET,
        data_description=[file, annotation, stream],
        session_names=SESSIONS,
        source_context=ctx,
        frame_size="1s",
        left_context="2s",
        right_context="2s",
        end="100s",
    )

    while (a := next(nova_iterator)):
        breakpoint()
        continue



import numpy as np
import decord
from decord import cpu
class DecordBatchReader(np.ndarray):
    """LazyArray class extending numpy.ndarray for video and audio loading."""
    import decord

    @property
    def shape(self):
        return self._shape

    def __init__(self, *args, **kwargs):
        super().__init__()

    def __new__(cls, filename):
        dr = decord.VideoReader(filename, ctx=cpu(0))
        _f = dr[0]
        dr.seek(0)

        shape = _f.shape[:2]
        dtype = np.dtype(_f.dtype)
        buffer = None

        obj = super().__new__(cls, shape[1:], dtype=dtype, buffer=buffer)
        obj.decord_reader = dr
        obj.start_idx = 0
        obj._num_samples = (shape[0])
        obj.len = dr._num_frame
        obj._shape = shape
        return obj

    def __len__(self):
        return self.shape[0]

    def __getitem__(self, index):
        if isinstance(index, slice):
            indices = list(range(index.start, index.stop))
            ret = self.decord_reader.get_batch(indices).asnumpy()
            # if type(self.decord_reader) == decord.video_reader.VideoReader:
            #     self.decord_reader.seek_accurate(index.start)
        elif isinstance(index, list):
            ret = np.squeeze(self.decord_reader.get_batch([index]).asnumpy())
            # if type(self.decord_reader) == decord.video_reader.VideoReader:
            #     self.decord_reader.seek(index[0])
        else:
            ret = self.decord_reader[index].asnumpy()
            # if type(self.decord_reader) == decord.video_reader.VideoReader:
            #     self.decord_reader.seek(index)
        return ret


"""discover_utils

Authors:
    Dominik Schiller <dominik.schiller@uni-a.de>
    Tobias Hallmen <tobias.hallmen@uni-a.de>
Date:
    06th September 2023

"""

_MAJOR_VERSION = '1'
_MINOR_VERSION = '0'
_PATCH_VERSION = '13'

__version__ = '.'.join([
    _MAJOR_VERSION,
    _MINOR_VERSION,
    _PATCH_VERSION,
])



"""Utility module to load and save xml files written by SSI

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    14.9.2023

"""
import xml.etree.ElementTree as ET
from pathlib import Path

from discover_utils.utils.string_utils import string_to_bool, parse_time_string_to_ms


class ModelIO:
    """
     ModelIO defines the inputs and outputs of a model in a chain or trainer file.

    This class is used to create and work with ChainLinks.

    Attributes:
        io_type (str): Defines if the object describes an input or an output. Can be either "input" or "output"
        io_id (str): String for the model to identify the input
        io_data(str): Description of the data <data_class>:<data_type>:<specific_data_type>

            ``"data_class"``
                     General data class. Either "annotation" or "stream"
            ``"data_type"``
                     Type of the data_class. Matches classnames data type from discover_utils.data .
                     "Video", "Audio", "SSIStream" for streams. "Discrete", "Free" or "Continuous" for annotations.
            ``"specific_data_type"``
                    Optional string identifier to specify either the annotation scheme or the type of feature. E.g. "transcript" or "Openface"

    Args:
        io_type (str): Defines if the object describes an input or an output. Can be either "input" or "output"
        io_id (str): String for the model to identify the input
        io_data(str): Description of the data <data_class>:<data_type>:<specific_data_type>

            ``"data_class"``
                     General data class. Either "annotation" or "stream"
            ``"data_type"``
                     Type of the data_class. Matches classnames data type from discover_utils.data .
                     "Video", "Audio", "SSIStream" for streams. "Discrete", "Free" or "Continuous" for annotations.
            ``"specific_data_type"``
                    Optional string identifier to specify either the annotation scheme or the type of feature. E.g. "transcript" or "Openface"
        io_default_value (str): Default value for stream name or annotation scheme to read from or write to
        io_default_active (bool): Only used for output streams. Default value to indicate if an output stream should be written to the output
    """

    def __init__(
            self,
            io_type: str,
            io_id: str,
            io_data: str,
            io_default_value: str,
            io_default_active: bool,
    ):
        """
        Initialize a ModelIO object with the specified parameters.

        """
        self.io_type = io_type
        self.io_id = io_id
        self.io_data = io_data
        self.io_default_value = io_default_value
        self.io_default_active = io_default_active


class URI:
    """
 URI objects are describing additional resources that are required by a module.

Attributes:
    uri_id (str): Defines if the object describes an input or an output. Can be either "input" or "output"
    uri_url (str): String for the model to identify the input
    uri_hash(str): Md5 hash to verify data integrity
    uri_tar(bool): If the uri describes a tarball

Args:
    uri_id (str): Defines if the object describes an input or an output. Can be either "input" or "output"
    uri_url (str, optional): String for the model to identify the input
    uri_hash(str, optional): Md5 hash to verify data integrity
    uri_tar(bool, optional): If the uri describes a tarball
"""

    def __init__(
            self,
            uri_id: str,
            uri_url: str = None,
            uri_hash: str = None,
            uri_tar: bool = False,
    ):
        """
        Initialize a ModelIO object with the specified parameters.

        """
        self.uri_id = uri_id
        self.uri_url = uri_url
        self.uri_hash = uri_hash
        self.uri_tar = uri_tar


class Trainer:
    """
    Class for representing and working with Trainer configuration.

    This class is used to create, load, and write Trainer configurations in XML format.

    Attributes:
        model_script_path (str): Path to the model script file.
        model_option_path (str): Path to the model option file.
        model_optstr (str): Model option string.
        model_weights_path (str): Path to the model weights file.
        model_stream (int): Model stream identifier.
        model_create (str): Model creation type.
        model_multi_role_input (bool): Indicates if the model accepts multi-role input.
        users (list): List of user configurations.
        classes (list): List of class configurations.
        streams (list): List of stream configurations.
        register (list): List of utilized dlls.
        info_trained (bool): Indicates if the Trainer has been trained.
        meta_frame_step (int): Frame step  for the Trainer.
        meta_right_ctx (int): Right context size for the Trainer.
        meta_left_ctx (int): Left context size for the Trainer.
        meta_balance (str): Balance type for the Trainer.
        meta_is_iterable (str): Bool that indicates if the module requires data processing via DISCOVER iterator.
        meta_enable_post_process (bool): Bool that indicates if the output specific postprocessing like packing or smoothing of the output should be enabled.
        meta_is_processable (str): Bool that indicates if the implements the Processor interface.
        meta_is_trainable (str): Bool that indicates if the implements the Trainer interface.
        meta_is_explainable (str): Bool that indicates if the module supports the Explain
        meta_backend (str): Backend of the model. E.g. sklearn, pytorch, tensorflow. Used to explain and train the model.
        meta_category (str): Category of the trainer.
        meta_description (str): Description of the trainer.
        meta_io(list[ModelIO]): Description of the inputs and outputs of the model.
        meta_uri(list[URI]): Description of additional resources required by the model.
        ssi_v (str): SSI version.
        xml_version (str): XML version.

    Args:
        model_script_path (str, optional): Path to the model script file. Defaults to empty string.
        model_option_path (str, optional): Path to the model option file. Defaults to empty string.
        model_option_string (str, optional): Model option string. Defaults to empty string.
        model_weights_path (str, optional): Path to the model weights file. Defaults to empty string.
        model_stream (int, optional): Model stream identifier. Default is 0.
        model_create (str, optional): Model creation type. Default is "PythonModel".
        model_multirole_input (bool, optional): Indicates if the model supports multi-role input. Default is False.
        users (list, optional): List of user information. Default is None.
        classes (list, optional): List of class information. Default is None.
        streams (list, optional): List of stream information. Default is None.
        register (list, optional): List of registered items. Default is None.
        info_trained (bool, optional): Indicates if the model is trained. Default is False.
        meta_is_iterable (str, optional): Bool that indicates if the module requires data processing via DISCOVER iterator. Defaults to False.
        meta_enable_post_process (bool, optional): Bool that indicates if the output specific postprocessing like packing or smoothing of the output should be enabled. Defaults to True.
        meta_is_processable (str, optional): Bool that indicates if the implements the Processor interface. Defaults to True.
        meta_is_trainable (str, optional): Bool that indicates if the implements the Trainer interface. Defaults to False.
        meta_is_explainable (str, optional): Bool that indicates if the module supports the Explain. Defaults to False.
        meta_frame_step (int, optional): Frame step value for metadata. Default is 0.
        meta_right_ctx (int, optional): Right context value for metadata. Default is 0.
        meta_left_ctx (int, optional): Left context value for metadata. Default is 0.
        meta_balance (str, optional): Balance type for metadata. Default is "none".

        meta_backend (str): Backend of the model. E.g. sklearn, pytorch, tensorflow. Used to explain and train the model. Defaults to 'unknown'.
        meta_category (str, optional): Category of the trainer. Default is "".
        meta_description (str, optional): Description of the trainer. Default is "".
        meta_io(list[ModelIO], optional): Description of the inputs and outputs of the model. Defaults to None.
        meta_uri(list[URI], optional): Description of additional resources required by the model.
        ssi_v (str, optional): SSI version. Default is "5".
        xml_version (str, optional): XML version. Default is "1.0".

    """

    def __init__(
            self,
            model_script_path: str = "",
            model_option_path: str = "",
            model_option_string: str = "",
            model_weights_path: str = "",
            model_stream: int = 0,
            model_create: str = "PythonModel",
            model_multirole_input=False,
            users: list = None,
            classes: list = None,
            streams: list = None,
            register: list = None,
            info_trained: bool = False,
            meta_frame_step: int = 0,
            meta_right_ctx: int = 0,
            meta_left_ctx: int = 0,
            meta_balance: str = "none",
            meta_backend: str = "unknown",
            meta_description: str = "",
            meta_category: str = "",
            meta_is_iterable: bool = False,
            meta_is_processable: bool = True,
            meta_is_trainable: bool = False,
            meta_is_explainable: bool = False,
            meta_enable_post_process: bool = True,
            meta_io: list[ModelIO] = None,
            meta_uri: list[URI] = None,
            ssi_v="5",
            xml_version="1.0",
    ):
        """
        Initialize a Trainer object with various parameters.

        """

        self.model_multi_role_input = None
        self.model_script_path = model_script_path
        self.model_option_path = model_option_path
        self.model_optstr = model_option_string
        self.model_weights_path = model_weights_path
        self.model_stream = model_stream
        self.model_create = model_create
        self.users = users if users is not None else []
        self.classes = classes if classes is not None else []
        self.streams = streams if streams is not None else []
        self.register = register if register is not None else []
        self.info_trained = info_trained
        self.meta_frame_step = meta_frame_step
        self.meta_right_ctx = meta_right_ctx
        self.meta_left_ctx = meta_left_ctx
        self.meta_balance = meta_balance
        self.meta_backend = meta_backend
        self.meta_description = meta_description
        self.meta_category = meta_category
        self.meta_is_trainable = meta_is_trainable
        self.meta_is_explainable = meta_is_explainable
        self.meta_is_iterable = meta_is_iterable
        self.meta_is_processable = meta_is_processable
        self.meta_enable_post_process = meta_enable_post_process
        self.meta_io = meta_io if meta_io is not None else []
        self.meta_uri = meta_uri if meta_uri is not None else []
        self.ssi_v = ssi_v
        self.xml_version = xml_version
        self.model_multi_role_input = model_multirole_input

    def load_from_file(self, fp):
        """
        Load Trainer configuration from an XML file.

        Args:
            fp (str or Path): The file path to the XML file.

        """
        root = ET.parse(Path(fp))
        info = root.find("info")
        meta = root.find("meta")
        register = root.find("register")
        streams = root.find("streams")
        classes = root.find("classes")
        users = root.find("users")
        model = root.find("model")

        if info is not None:
            self.info_trained = string_to_bool(info.get("trained", ""))
        if meta is not None:
            self.meta_left_ctx = parse_time_string_to_ms(meta.get("leftContext", default=self.meta_left_ctx),
                                                         suppress_warn=True)
            self.meta_right_ctx = parse_time_string_to_ms(meta.get("rightContext", default=self.meta_right_ctx),
                                                          suppress_warn=True)
            self.meta_frame_step = parse_time_string_to_ms(meta.get("frameStep", default=self.meta_frame_step),
                                                           suppress_warn=True)
            self.meta_balance = meta.get("balance", default=self.meta_balance)
            self.meta_backend = meta.get("backend", default=self.meta_backend)
            self.meta_description = meta.get("description", default=self.meta_description)
            self.meta_category = meta.get("category", default=self.meta_category)
            self.meta_is_iterable = string_to_bool(meta.get("is_iterable", default=self.meta_is_iterable))
            self.meta_is_processable = string_to_bool(meta.get("is_processable", default=self.meta_is_processable))
            self.meta_is_trainable = string_to_bool(meta.get("is_trainable", default=self.meta_is_trainable))
            self.meta_is_explainable = string_to_bool(meta.get("is_explainable", default=self.meta_is_explainable))
            self.meta_is_explainable = string_to_bool(meta.get("is_explainable", default=self.meta_is_explainable))
            self.meta_enable_post_process = string_to_bool(meta.get("enable_post_process", default=self.meta_enable_post_process))

            for io_tag in meta.findall("io"):
                self.meta_io.append(
                    ModelIO(io_tag.get("type"), io_tag.get("id"), io_tag.get("data"), io_tag.get("default_value"), string_to_bool(io_tag.get("default_active", default="True")))
                )

            for uri_tag in meta.findall("uri"):
                self.meta_uri.append(
                    URI(uri_tag.get("id"), uri_tag.get("url"), uri_tag.get("hash"),
                        string_to_bool(uri_tag.get("tar", "")))
                )
        if register is not None:
            for r in register:
                self.register.append(r.attrib)
        if streams is not None:
            for s in streams:
                self.streams.append(s.attrib)
        if classes is not None:
            for c in classes:
                self.classes.append(c.attrib)
        if users is not None:
            for u in users:
                self.users.append(u.attrib)
        if model is not None:
            self.model_stream = model.get("stream", "0")
            self.model_create = model.get("create", "PythonModel")
            self.model_option_path = model.get("option", "")
            self.model_script_path = model.get("script", "")
            self.model_weights_path = model.get("path", "")
            self.model_optstr = model.get("optstr", "")
            self.model_multi_role_input = string_to_bool(model.get("multi_role_input", ""))

    def write_to_file(self, fp):
        """
        Write Trainer configuration to an XML file.

        Args:
            fp (str or Path): The file path to save the XML file.

        """
        root = ET.Element("trainer")
        ET.SubElement(root, "info", trained=str(self.info_trained))
        meta = ET.SubElement(
            root,
            "meta",
            frameStep=str(self.meta_frame_step),
            leftContext=str(self.meta_left_ctx),
            rightContex=str(self.meta_right_ctx),
            balance=self.meta_balance,
            backend=self.meta_backend,
            category=self.meta_category,
            description=self.meta_description,
            meta_is_iterable=str(self.meta_is_iterable),
            meta_is_trainable=str(self.meta_is_trainable),
            meta_is_explainable=str(self.meta_is_explainable),
            meta_enable_post_process=str(self.meta_enable_post_process)
        )

        io: ModelIO
        for io in self.meta_io:
            ET.SubElement(
                meta,
                "io",
                id=io.io_id,
                type=io.io_type,
                data=io.io_data,
                default_value=io.io_default_value
            )

        uri: URI
        for uri in self.meta_uri:
            ET.SubElement(
                meta,
                "uri",
                id=uri.uri_id,
                url=uri.uri_url,
                hash=uri.uri_hash,
                tar=str(uri.uri_tar)
            )

        register = ET.SubElement(root, "register")
        for r in self.register:
            ET.SubElement(register, "item", **r)
        streams = ET.SubElement(root, "streams")
        for s in self.streams:
            ET.SubElement(streams, "item", **s)
        classes = ET.SubElement(root, "classes")
        for c in self.classes:
            ET.SubElement(classes, "item", **c)
        users = ET.SubElement(root, "users")
        for u in self.users:
            ET.SubElement(users, "item", **u)
        ET.SubElement(
            root,
            "model",
            create=self.model_create,
            stream=str(self.model_stream),
            path=self.model_weights_path,
            script=self.model_script_path,
            optstr=self.model_optstr,
            option=self.model_option_path,
        )

        tree = ET.ElementTree(root)
        ET.indent(tree, space="    ", level=0)

        if not fp.suffix:
            fp = fp.with_suffix(".trainer")
        tree.write(fp)


class ChainLink:
    """
    Class for representing single steps of a SSI chain configuration.

    This class is used to create and work with ChainLinks.

    Attributes:
        create (str): ChainLink type.
        script (str): ChainLink script.
        optsstr (str): ChainLink options string.
        syspath (str): ChainLink system path.
        tag (str): ChainLink tag.
        multi_role_input (bool): Indicates if the ChainLink accepts multi-role input.

    Args:
        create (str): ChainLink type.
        script (str): ChainLink script.
        optsstr (str): ChainLink options string.
        syspath (str): ChainLink system path.
        tag (str, optional): ChainLink tag. Defaults tu "feature"
        multi_role_input (str, optional): Indicates if the ChainLink accepts multi-role input. Defaults to False.

    """

    def __init__(
            self,
            create: str,
            script: str,
            optsstr: str,
            syspath: str,
            tag: str = "feature",
            multi_role_input: str = "False",
            **kwargs,
    ):
        """
        Initialize a ChainLink object with the specified parameters.

        """
        self.create = create
        self.script = script
        self.optsstr = optsstr
        self.syspath = syspath
        self.tag = tag
        self.multi_role_input = True if multi_role_input == "True" else False


class Chain:
    """
    Class for representing and working with Chain configuration.

    This class is used to create, load, and write Chain configurations in XML format.

    Attributes:
        meta_frame_step (str): Meta frame step value.
        meta_left_ctx (str): Meta left context value.
        meta_right_ctx (str): Meta right context value.
        meta_backend (str): Backend type for the Chain.
        meta_description (str): Description for the Chain.
        meta_category (str): Category for the Chain.
        meta_io(list[ModelIO], optional): Description of the inputs and outputs of the model.
        register (list): List of register configurations.
        links (list): List of ChainLink configurations.

    Args:
        meta_frame_step (str, optional): Meta frame step information. Defaults to empty string.
        meta_left_context (str, optional): Left context metadata. Defaults to empty string.
        meta_right_context (str, optional): Right context metadata. Defaults to empty string.
        meta_backend (str, optional): Backend type for metadata. Default is "nova-server".
        meta_description (str, optional): Description for metadata. Defaults to empty string.
        meta_category (str, optional): Category for metadata. Defaults to empty string.
        meta_io(list[ModelIO], optional): Description of the inputs and outputs of the model. Defaults to None.
        register (list, optional): List of registered items. Default is None.
        links (list, optional): List of ChainLink objects. Default is None.

    """

    def __init__(
            self,
            meta_frame_step: str = "",
            meta_left_context: str = "",
            meta_right_context: str = "",
            meta_backend: str = "nova-server",
            meta_description: str = "",
            meta_category: str = "",
            meta_io: list[ModelIO] = None,
            register: list = None,
            links: list = None,
    ):
        """
        Initialize a Chain object with the specified parameters.

        """
        self.meta_frame_step = meta_frame_step
        self.meta_left_ctx = meta_left_context
        self.meta_right_ctx = meta_right_context
        self.meta_backend = meta_backend
        self.meta_description = meta_description
        self.meta_category = meta_category
        self.meta_io = meta_io if meta_io is not None else []
        self.register = register if register else []
        self.links = links if links else []

    def load_from_file(self, fp):
        """
        Load Chain configuration from an XML file.

        Args:
            fp (str or Path): The file path to the XML file.

        """
        tree = ET.parse(Path(fp))
        root = tree.getroot()
        meta = tree.find("meta")
        register = tree.find("register")
        links = []
        for child in root:
            if child.tag == "feature" or child.tag == "filter":
                links.append(child)

        if meta is not None:
            self.meta_frame_step = meta.attrib.get("frameStep", "0")
            self.meta_left_ctx = meta.attrib.get("leftContext", "0")
            self.meta_right_ctx = meta.attrib.get("rightContext", "0")
            self.meta_backend = meta.attrib.get("backend", "nova-server")
            self.meta_description = meta.attrib.get("description", "")
            self.meta_category = meta.attrib.get("category", "")
            for io_tag in meta.findall("io"):
                self.meta_io.append(
                    ModelIO(io_tag.get("type"), io_tag.get("id"), io_tag.get("data"), io_tag.get("default_value"))
                )

        if register is not None:
            for r in register:
                self.register.append(r.attrib)

        for link in links:
            item = link.find("item")
            new_link = ChainLink(**item.attrib, tag=link.tag)
            self.links.append(new_link)

    def write_to_file(self, fp):
        """
        Write Chain configuration to an XML file.

        Args:
            fp (str or Path): The file path to save the XML file.

        """
        root = ET.Element("chain")
        meta = ET.SubElement(
            root,
            "meta",
            frameStep=str(self.meta_frame_step),
            leftContext=str(self.meta_left_ctx),
            rightContex=str(self.meta_right_ctx),
            backend=str(self.meta_backend),
            description=str(self.meta_description),
            category=str(self.meta_category),
        )

        io: ModelIO
        for io in self.meta_io:
            ET.SubElement(
                meta,
                "io",
                id=io.io_id,
                type=io.io_type,
                data=io.io_data
            )

        register = ET.SubElement(root, "register")
        for r in self.register:
            ET.SubElement(register, "item", **r)

        cl: ChainLink
        for cl in self.links:
            link = ET.SubElement(root, cl.tag)
            ET.SubElement(
                link,
                "item",
                create=cl.create,
                script=cl.script,
                syspath=cl.syspath,
                optsstr=cl.optsstr,
                multi_role_input=str(cl.multi_role_input),
            )

        tree = ET.ElementTree(root)
        ET.indent(tree, space="    ", level=0)

        if not fp.suffix:
            fp = fp.with_suffix(".chain")
        tree.write(fp)


if __name__ == "__main__":
    from pathlib import Path
    import os
    import dotenv
    dotenv.load_dotenv('../.env')
    data_dir = Path(os.getenv("DISCOVER_DATA_DIR"))
    out_dir = Path(os.getenv("DISCOVER_TEST_DIR"))

    trainer_in_fp = data_dir / "test.trainer"
    trainer_out_fp = out_dir / "test_trainer.trainer"

    trainer = Trainer()
    trainer.load_from_file(trainer_in_fp)
    trainer.write_to_file(trainer_out_fp)

    # chain_in_fp = Path('')
    # chain_out_fp = Path("test_chain.chain")
    #
    # chain = Chain()
    # chain.load_from_file(chain_in_fp)
    # chain.write_to_file(chain_out_fp)
    breakpoint()



import typing
from abc import ABC, abstractmethod
from time import perf_counter
from discover_utils.data.annotation import Annotation
from discover_utils.data.provider.dataset_iterator import DatasetIterator
from discover_utils.data.provider.data_manager import DatasetManager, SessionManager
from discover_utils.data.stream import Stream
from discover_utils.utils.log_utils import log
from discover_utils.utils.ssi_xml_utils import ModelIO
from discover_utils.utils.ssi_xml_utils import Trainer as SSITrainer


class Processor(ABC):
    """
    Base class of a data processor. This interface builds the foundation for all data processing classes.
    """

    # List of dependencies that need to be installed when the script is loaded
    DEPENDENCIES = []

    # Flag to indicate whether the processed input belongs to one role or to multiple roles
    SINGLE_ROLE_INPUT = True

    # Prefix for provided datastreams that are missing the id-tag
    UNKNOWN_ID = '<unk>_'

    # TODO read trainer or chain file for default options
    def __init__(self, model_io: list[ModelIO], opts: dict, trainer: SSITrainer = None):
        self.model = None
        self.data = None
        self.output = None
        self.options = opts
        self.model_io = model_io
        self.trainer = trainer
        self.ds_manager = None

    def get_session_manager(self, dataset_manager: DatasetManager, session_name: str = None, ignore_ambiguity: bool = False) -> SessionManager:
        """Returns a session manager that matches the provides session name. If no session name is provided the first

        Args:
            dataset_manager (DatasetManager): The dataset manager that holds all sessions
            session_name (str, Optional): The name of a specific session to retrieve. Defaults to None.
            ignore_ambiguity (bool, Optional): If set to False and no session name is specified and more than one session exists the function throws an error. Else the first session in the dictionary will be returned.

        Raises:
            ValueError: If session to return cannot be identified
        """
        current_session_name = session_name
        if current_session_name is None:
            if len(dataset_manager.session_names) == 1 or ignore_ambiguity:
                current_session_name = dataset_manager.session_names[0]
            else:
                raise ValueError('')

        return dataset_manager.sessions[current_session_name]['manager']

    def preprocess_sample(self, sample: dict):
        """Preprocess data to convert between DISCOVER dataset iterator item to the raw model input as required in process_sample.

        Args:
            sample :
        """
        return list(sample.values())[0]

    def process_sample(self, sample):
        """Applying processing steps (e.g. feature extraction, data prediction etc... ) to the provided data."""
        return sample

    def postprocess_sample(self, sample):
        """Apply any optional postprocessing to the data (e.g. scaling, mapping etc...)"""
        return sample

    def process_data(self, ds_manager: DatasetManager) -> dict:
        """Returning a dictionary that contains the original keys from the dataset iterator and a list of processed
        samples as value. Can be overwritten to customize the processing"""
        ds_manager: DatasetIterator
        self.ds_manager = ds_manager
        # Start the stopwatch / counter
        pc_start = perf_counter()
        output_list = []
        for i, sample in enumerate(ds_manager):
            if i % 100 == 0:
                log(f'Processing sample {i}. {i / (perf_counter() - pc_start)} samples / s. Processed {i * ds_manager.stride / 1000} Seconds of data.')
            # for id, output_list in processed.items():
            #     data_for_id = {id: sample[id]}
            out = self.preprocess_sample(sample)
            out = self.process_sample(out)
            out = self.postprocess_sample(out)
            output_list.append(out)


        # TODO convert output to dict
        return output_list

    def to_output(self, data):
        if isinstance(self, Predictor):
            return self.to_anno(data)
        elif isinstance(self, Extractor):
            return self.to_stream(data)

class Trainer(ABC):
    """
    Base class of a Trainer. Implement this interface in your own class to build a model that is trainable from within nova
    """

    """Includes all the necessary files to run this script"""

    @abstractmethod
    def train(self):
        """Trains a model with the given data."""
        raise NotImplemented

    @abstractmethod
    def save(self, path) -> str:
        """Stores the weights of the given model at the given path. Returns the path of the weights."""
        raise NotImplemented

    @abstractmethod
    def load(self, path):
        """Loads a model with the given path. Returns this model."""
        raise NotImplemented

class Explainer(ABC):
    @abstractmethod
    def get_explainable_model(self) -> typing.Any:
        raise NotImplemented
    @abstractmethod
    def get_predict_function(self) -> typing.Callable:
        raise NotImplemented

class Predictor(Processor):
    """
    Base class of a data predictor. Implement this interface if you want to write annotations to a database
    """

    @abstractmethod
    def to_anno(self, data) -> list[Annotation]:
        """Converts the output of process_data to the correct annotation format to upload them to the database.
        !THE OUTPUT FORMAT OF THIS FUNCTION IS NOT YET FULLY DEFINED AND WILL CHANGE IN FUTURE RELEASES!

        Args:
            data (object): Data output of process_data function

        Returns:
            list: A list of annotation objects
        """
        raise NotImplemented


class Extractor(Processor):
    """
    Base class of a feature extractor. Implement this interface in your own class to build a feature extractor.
    """

    @property
    @abstractmethod
    def chainable(self):
        """Whether this extraction module can be followed by other extractors. If set to True 'to_ds_iterable()' must be implemented"""
        return False

    @abstractmethod
    def to_stream(self, data: object) -> list[Stream]:
        """Converts the return value from process_data() to data stream chunk that can be processed by DISCOVER.

        Args:
            data (object): The data as returned by the process_data function of the Processor class


        Returns:
            list: A list of stream objects.

        """
        raise NotImplemented



"""Utilities for file download and caching. Partially transferred from https://github.com/tensorflow/tensorflow/blob/v2.5.0/tensorflow/python/keras/utils/data_utils.py#L148-L277"""
import hashlib
import os
import shutil
import tarfile
import tempfile
import urllib
import zipfile
import requests
import mimetypes
from tempfile import NamedTemporaryFile
from pathlib import Path
from urllib import request, error
from zipfile import ZipFile

def retreive_from_url(url, fp=None):

    with requests.get(url, stream=True, headers={'Accept-Encoding': None}) as r:

        if fp is None:
            suffix = mimetypes.guess_extension(r.headers.get('content-type'))
            tmp_file = tempfile.NamedTemporaryFile(suffix=suffix, delete=False)
            fp = Path(tmp_file.name)
        # save the output to a file
        with open(fp, 'wb')as output:
            shutil.copyfileobj(r.raw, output)

        return fp, r.headers
def _resolve_hasher(algorithm, file_hash=None):
    """Returns hash algorithm as hashlib function."""
    if algorithm == "sha256":
        return hashlib.sha256()

    if algorithm == "auto" and file_hash is not None and len(file_hash) == 64:
        return hashlib.sha256()

    # This is used only for legacy purposes.
    return hashlib.md5()


def _hash_file(fpath, algorithm="sha256", chunk_size=65535):
    """Calculates a file sha256 or md5 hash.

    Example:

    ```python
    _hash_file('/path/to/file.zip')
    'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    ```

    Args:
        fpath: path to the file being validated
        algorithm: hash algorithm, one of `'auto'`, `'sha256'`, or `'md5'`.
            The default `'sha256'` uses sha256.
            Note: when 'auto' is passed as a string (not a hasher object),
            it will default to md5 for legacy compatibility.
        chunk_size: Bytes to read at a time, important for large files.

    Returns:
        The file hash
    """
    if isinstance(algorithm, str):
        # Note: we don't have file_hash here, so 'auto' will default to md5
        hasher = _resolve_hasher(algorithm)
    else:
        hasher = algorithm

    with open(fpath, "rb") as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b""):
            hasher.update(chunk)

    return hasher.hexdigest()


def _extract_archive(file_path, path=".", archive_format="auto"):
    """Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.

    Args:
        file_path: path to the archive file
        path: path to extract the archive file
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    Returns:
        True if a match was found and an archive extraction was completed,
        False otherwise.
    """
    if archive_format is None:
        return False
    if archive_format == "auto":
        archive_format = ["tar", "zip"]
    if isinstance(archive_format, str):
        archive_format = [archive_format]

    file_path = file_path
    path = path

    for archive_type in archive_format:
        if archive_type == "tar":
            open_fn = tarfile.open
            is_match_fn = tarfile.is_tarfile
        if archive_type == "zip":
            open_fn = zipfile.ZipFile
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError, KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False

def retreive_and_unzip(url, extract_to='./tmp', tmp_dir='./tmp'):
    fn = url.split('/')[-1]
    tmp_dir = Path(tmp_dir)
    tmp_dir.mkdir(parents=True, exist_ok=True)

    tmp_file = tmp_dir / fn

    if tmp_file.is_file():
        print(f'File {tmp_file} already exists. Skipping download.')
    else:
        retreive_from_url(url, tmp_file)

    with ZipFile(tmp_file) as zObject:
        zObject.extractall(path=Path(extract_to))


def validate_file(fpath, file_hash, algorithm="auto", chunk_size=65535):
    """Validates a file against a sha256 or md5 hash.

    Args:
        fpath: path to the file being validated
        file_hash:  The expected hash string of the file.
            The sha256 and md5 hash algorithms are both supported.
        algorithm: Hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    Returns:
        Whether the file is valid
    """
    # Normalize the expected hash by stripping whitespace and converting to lowercase
    file_hash = str(file_hash).strip().lower()

    hasher = _resolve_hasher(algorithm, file_hash)

    if str(_hash_file(fpath, hasher, chunk_size)) == str(file_hash):
        return True
    else:
        return False


def get_file(
        fname,
        origin,
        untar=False,
        md5_hash=None,
        file_hash=None,
        cache_subdir="downloads",
        hash_algorithm="auto",
        extract=False,
        archive_format="auto",
        cache_dir=None,
        tmp_dir=None
):
    """Downloads a file from a URL if it not already in the cache.

    By default, the file at the url `origin` is downloaded to the
    cache_dir '~/.hcai_downloads', placed in the cache_subdir `downloads`,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `~/.hcai_downloads/downloads/example.txt`.

    Files in tar, tar.gz, tar.bz, and zip formats can also be extracted.
    Passing a hash will verify the file after download. The command line
    programs `shasum` and `sha256sum` can compute the hash.

    Example:

    ```python
    #TODO
    ```

    Args:
        fname: Name of the file. If an absolute path `/path/to/file.txt` is
            specified the file will be saved at that location.
        origin: Original URL of the file.
        untar: Deprecated in favor of `extract` argument.
            boolean, whether the file should be decompressed
        md5_hash: Deprecated in favor of `file_hash` argument.
            md5 hash of the file for verification
        file_hash: The expected hash string of the file after download.
            The sha256 and md5 hash algorithms are both supported.
        cache_subdir: Subdirectory under the cache dir where the file is
            saved. If an absolute path `/path/to/folder` is
            specified the file will be saved at that location.
        hash_algorithm: Select the hash algorithm to verify the file.
            options are `'md5'`, `'sha256'`, and `'auto'`.
            The default 'auto' detects the hash algorithm in use.
        extract: True tries extracting the file as an Archive, like tar or zip.
        archive_format: Archive format to try for extracting the file.
            Options are `'auto'`, `'tar'`, `'zip'`, and `None`.
            `'tar'` includes tar, tar.gz, and tar.bz files.
            The default `'auto'` corresponds to `['tar', 'zip']`.
            None or an empty list will return no matches found.
        cache_dir: Location to store cached files, when None it
            defaults to the default directory `~/.nova_utils/`.
        tmp_dir: Location to download temporary files, when no tmp_dir is specificed
            cache dir is used.

    Returns:
        Path to the downloaded file
    """
    if cache_dir is None:
        cache_dir = Path.home() / ".nova_utils"
    else:
        cache_dir = Path(cache_dir)

    if tmp_dir is None:
        tmp_dir = Path(cache_dir)
    else:
        tmp_dir = Path(tmp_dir)

    if md5_hash is not None and file_hash is None:
        file_hash = md5_hash
        hash_algorithm = "md5"

    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    if not os.access(cache_dir, os.W_OK):
        cache_dir = os.path.join("/tmp", ".hcai_models")

    data_dir = cache_dir / cache_subdir
    data_dir.mkdir(parents=True, exist_ok=True)

    #TODO if a zipfile is loaded we only compare the hash of the zip file. If the original zip file is deleted it
    # will be downloaded again every time

    if untar:
        untar_fpath = data_dir / fname
        fpath = untar_fpath.parent / (untar_fpath.name + '.tar.gz')
    else:
        fpath = data_dir / fname

    download = False
    if fpath.exists():
        # File found; verify integrity if a hash was provided.
        if file_hash:
            if not validate_file(fpath, file_hash, algorithm=hash_algorithm):
                print(
                    "A local file was found, but it seems to be "
                    "incomplete or outdated because the "
                    + hash_algorithm
                    + " file hash does not match the original value of "
                    + file_hash
                    + " so we will re-download the data."
                )
                download = True
    else:
        download = True

    if download:
        print("Downloading data from", origin)

        error_msg = "URL fetch failure on {}: {} -- {}"
        try:
            try:
                tmp_path = tmp_dir / fname
                #request.urlretrieve(origin, tmp_path)
                retreive_from_url(origin, fp=tmp_path)
                shutil.move(tmp_path, fpath)
            except urllib.error.HTTPError as e:
                raise Exception(error_msg.format(origin, e.code, str(e)))
            except urllib.error.URLError as e:
                raise Exception(error_msg.format(origin, e.errno, e.reason))
        except (Exception, KeyboardInterrupt) as e:
            if os.path.exists(fpath):
                os.remove(fpath)
            raise

    if untar:
        if not untar_fpath.exists():
            _extract_archive(fpath, untar_fpath, archive_format="tar")
            #fpath.unlink()
        return untar_fpath

    if extract:
        _extract_archive(fpath, data_dir, archive_format)

    return fpath


if __name__ == '__main__':
    from pathlib import Path
    import os
    import dotenv
    dotenv.load_dotenv()
    data_dir = Path(os.getenv("DISCOVER_DATA_DIR"))
    cache_dir = Path(os.getenv("DISCOVER_CACHE_DIR"))
    tmp_dir = Path(os.getenv("DISCOVER_TMP_DIR"))
    pth_url = os.getenv("DISCOVER_TEST_PTH_URL")
    st_url = os.getenv("DISCOVER_TEST_ST_URL")

    get_file(fname='test.safetensor', origin=st_url)

    # Download file and get filepath
    local_fp = get_file(fname=data_dir/'blazeface.pth', origin=pth_url)

    # Untar local file
    local_fp = get_file(fname=data_dir/'test', origin=pth_url, untar=True)

    # Unzip local file
    local_fp = get_file(fname=data_dir/'blazeface_test.pth.zip', origin=pth_url, extract=True)

